"""Command models for opening a gripper jaw."""
from __future__ import annotations
from typing import Literal, Type, Optional, Any

from pydantic import BaseModel, Field
from pydantic.json_schema import SkipJsonSchema

from opentrons.hardware_control import HardwareControlAPI
from opentrons.protocol_engine.resources import ensure_ot3_hardware

from ..command import (
    AbstractCommandImpl,
    BaseCommand,
    BaseCommandCreate,
    SuccessData,
)
from opentrons.protocol_engine.errors.error_occurrence import ErrorOccurrence


closeGripperJawCommandType = Literal["robot/closeGripperJaw"]


def _remove_default(s: dict[str, Any]) -> None:
    s.pop("default", None)


class closeGripperJawParams(BaseModel):
    """Payload required to close a gripper."""

    force: float | SkipJsonSchema[None] = Field(
        default=None,
        description="The force the gripper should use to hold the jaws, falls to default if none is provided.",
        json_schema_extra=_remove_default,
    )


class closeGripperJawResult(BaseModel):
    """Result data from the execution of a closeGripperJaw command."""

    pass


class closeGripperJawImplementation(
    AbstractCommandImpl[closeGripperJawParams, SuccessData[closeGripperJawResult]]
):
    """closeGripperJaw command implementation."""

    def __init__(
        self,
        hardware_api: HardwareControlAPI,
        **kwargs: object,
    ) -> None:
        self._hardware_api = hardware_api

    async def execute(
        self, params: closeGripperJawParams
    ) -> SuccessData[closeGripperJawResult]:
        """Release the gripper."""
        ot3_hardware_api = ensure_ot3_hardware(self._hardware_api)
        await ot3_hardware_api.grip(force_newtons=params.force)
        return SuccessData(
            public=closeGripperJawResult(),
        )


class closeGripperJaw(
    BaseCommand[closeGripperJawParams, closeGripperJawResult, ErrorOccurrence]
):
    """closeGripperJaw command model."""

    commandType: closeGripperJawCommandType = "robot/closeGripperJaw"
    params: closeGripperJawParams
    result: Optional[closeGripperJawResult] = None

    _ImplementationCls: Type[
        closeGripperJawImplementation
    ] = closeGripperJawImplementation


class closeGripperJawCreate(BaseCommandCreate[closeGripperJawParams]):
    """closeGripperJaw command request model."""

    commandType: closeGripperJawCommandType = "robot/closeGripperJaw"
    params: closeGripperJawParams

    _CommandCls: Type[closeGripperJaw] = closeGripperJaw



"""Executor for liquid class based complex commands."""
from __future__ import annotations

import logging
from copy import deepcopy
from enum import Enum
from typing import TYPE_CHECKING, Optional, Union, Literal
from dataclasses import dataclass, field

from opentrons_shared_data.liquid_classes.liquid_class_definition import (
    PositionReference,
    Coordinate,
    BlowoutLocation,
)
from opentrons_shared_data.pipette.types import LIQUID_PROBE_START_OFFSET_FROM_WELL_TOP

from opentrons.protocol_api._liquid_properties import (
    Submerge,
    TransferProperties,
    MixProperties,
    SingleDispenseProperties,
    MultiDispenseProperties,
    TouchTipProperties,
)
from opentrons.protocol_engine.errors import TouchTipDisabledError
from opentrons.types import Location, Point
from opentrons.protocols.advanced_control.transfers.transfer_liquid_utils import (
    LocationCheckDescriptors,
)
from opentrons.protocols.advanced_control.transfers import (
    transfer_liquid_utils as tx_utils,
)

if TYPE_CHECKING:
    from .well import WellCore
    from .instrument import InstrumentCore
    from ... import TrashBin, WasteChute

log = logging.getLogger(__name__)


@dataclass
class LiquidAndAirGapPair:
    """Pairing of a liquid and air gap in a tip, with air gap below the liquid in a tip."""

    liquid: float = 0
    air_gap: float = 0


@dataclass
class TipState:
    """Carrier of the state of the pipette tip in use.

    Properties:
        last_liquid_and_air_gap_in_tip: The last liquid + air_gap combo in the tip.
            This will only include the existing liquid and air gap in the tip that
            an aspirate/ dispense interacts with. For example, the air gap from
            a previous step that needs to be removed, or the liquid from a previous
            aspirate that needs to be dispensed or the liquid that needs to be added to
            during a consolidation.
        ready_to_aspirate: Whether the pipette plunger is in a position that allows
            correct aspiration. The starting state for the pipette at initialization of
            `TransferComponentsExecutor`s should be ready_to_aspirate == True.
    """

    ready_to_aspirate: bool = True
    # TODO: maybe use the tip contents from engine state instead.
    last_liquid_and_air_gap_in_tip: LiquidAndAirGapPair = field(
        default_factory=LiquidAndAirGapPair
    )

    def append_liquid(self, volume: float) -> None:
        # Neither aspirate nor a dispense process should be adding liquid
        # when there is an air gap present.
        assert (
            self.last_liquid_and_air_gap_in_tip.air_gap == 0
        ), "Air gap present in the tip."
        self.last_liquid_and_air_gap_in_tip.liquid += volume

    def delete_liquid(self, volume: float) -> None:
        # Neither aspirate nor a dispense process should be removing liquid
        # when there is an air gap present.
        assert (
            self.last_liquid_and_air_gap_in_tip.air_gap == 0
        ), "Air gap present in the tip."
        self.last_liquid_and_air_gap_in_tip.liquid -= volume

    def append_air_gap(self, volume: float) -> None:
        # Neither aspirate nor a dispense process should be adding air gaps
        # when there is already an air gap present.
        assert (
            self.last_liquid_and_air_gap_in_tip.air_gap == 0
        ), "Air gap already present in the tip."
        self.last_liquid_and_air_gap_in_tip.air_gap = volume

    def delete_air_gap(self, volume: float) -> None:
        assert (
            self.last_liquid_and_air_gap_in_tip.air_gap == volume
        ), "Last air gap volume doe not match the volume being removed"
        self.last_liquid_and_air_gap_in_tip.air_gap = 0


class TransferType(Enum):
    ONE_TO_ONE = "one_to_one"
    MANY_TO_ONE = "many_to_one"
    ONE_TO_MANY = "one_to_many"


class TransferComponentsExecutor:
    def __init__(
        self,
        instrument_core: InstrumentCore,
        transfer_properties: TransferProperties,
        target_location: Location,
        target_well: WellCore,
        tip_state: TipState,
        transfer_type: TransferType,
    ) -> None:
        self._instrument = instrument_core
        self._transfer_properties = transfer_properties
        self._target_location = target_location
        self._target_well = target_well
        self._tip_state: TipState = deepcopy(tip_state)  # don't modify caller's object
        self._transfer_type: TransferType = transfer_type

    @property
    def tip_state(self) -> TipState:
        """Return the tip state."""
        return self._tip_state

    def submerge(
        self,
        submerge_properties: Submerge,
        post_submerge_action: Literal["aspirate", "dispense"],
        volume_for_pipette_mode_configuration: Optional[float],
    ) -> None:
        """Execute submerge steps.

        1. move to position shown by positionReference + offset (should practically be a point outside/above the liquid).
        Should raise an error if this point is inside the liquid?
            For liquid meniscus this is easy to tell. Can’t be below meniscus
            For reference pos of anything else, do not allow submerge position to be below aspirate position
        2. move to aspirate position at desired speed
        3. delay
        """
        submerge_start_point = absolute_point_from_position_reference_and_offset(
            well=self._target_well,
            position_reference=submerge_properties.position_reference,
            offset=submerge_properties.offset,
        )
        submerge_start_location = Location(
            point=submerge_start_point, labware=self._target_location.labware
        )
        prep_before_moving_to_submerge = (
            post_submerge_action == "aspirate"
            and volume_for_pipette_mode_configuration is not None
        )
        if prep_before_moving_to_submerge:
            # Move to the tip probe start position
            self._instrument.move_to(
                location=Location(
                    point=self._target_well.get_top(
                        LIQUID_PROBE_START_OFFSET_FROM_WELL_TOP.z
                    ),
                    labware=self._target_location.labware,
                ),
                well_core=self._target_well,
                force_direct=False,
                minimum_z_height=None,
                speed=None,
            )
            self._remove_air_gap(location=submerge_start_location)
            if (
                self._transfer_type != TransferType.MANY_TO_ONE
                and self._instrument.get_liquid_presence_detection()
            ):
                self._instrument.liquid_probe_with_recovery(
                    well_core=self._target_well, loc=submerge_start_location
                )
            # TODO: do volume configuration + prepare for aspirate only if the mode needs to be changed
            self._instrument.configure_for_volume(volume_for_pipette_mode_configuration)  # type: ignore[arg-type]
            self._instrument.prepare_to_aspirate()

        tx_utils.raise_if_location_inside_liquid(
            location=submerge_start_location,
            well_location=self._target_location,
            well_core=self._target_well,
            location_check_descriptors=LocationCheckDescriptors(
                location_type="submerge start",
                pipetting_action=post_submerge_action,
            ),
            logger=log,
        )
        self._instrument.move_to(
            location=submerge_start_location,
            well_core=self._target_well,
            force_direct=False,
            minimum_z_height=None,
            speed=None,
        )
        if not prep_before_moving_to_submerge:
            self._remove_air_gap(location=submerge_start_location)
        self._instrument.move_to(
            location=self._target_location,
            well_core=self._target_well,
            force_direct=True,
            minimum_z_height=None,
            speed=submerge_properties.speed,
        )
        if submerge_properties.delay.enabled:
            assert submerge_properties.delay.duration is not None
            self._instrument.delay(submerge_properties.delay.duration)

    def aspirate_and_wait(self, volume: float) -> None:
        """Aspirate according to aspirate properties and wait if enabled."""
        # TODO: handle volume correction
        aspirate_props = self._transfer_properties.aspirate
        correction_volume = aspirate_props.correction_by_volume.get_for_volume(volume)
        self._instrument.aspirate(
            location=self._target_location,
            well_core=None,
            volume=volume,
            rate=1,
            flow_rate=aspirate_props.flow_rate_by_volume.get_for_volume(volume),
            in_place=True,
            correction_volume=correction_volume,
        )
        self._tip_state.append_liquid(volume)
        delay_props = aspirate_props.delay
        if delay_props.enabled:
            # Assertion only for mypy purposes
            assert delay_props.duration is not None
            self._instrument.delay(delay_props.duration)

    def dispense_and_wait(
        self,
        dispense_properties: Union[SingleDispenseProperties, MultiDispenseProperties],
        volume: float,
        push_out_override: Optional[float],
    ) -> None:
        """Dispense according to dispense properties and wait if enabled."""
        correction_volume = dispense_properties.correction_by_volume.get_for_volume(
            volume
        )
        self._instrument.dispense(
            location=self._target_location,
            well_core=None,
            volume=volume,
            rate=1,
            flow_rate=dispense_properties.flow_rate_by_volume.get_for_volume(volume),
            in_place=True,
            push_out=push_out_override,
            correction_volume=correction_volume,
        )
        if push_out_override:
            # If a push out was performed, we need to reset the plunger before we can aspirate again
            self._tip_state.ready_to_aspirate = False
        self._tip_state.delete_liquid(volume)
        dispense_delay = dispense_properties.delay
        if dispense_delay.enabled:
            assert dispense_delay.duration is not None
            self._instrument.delay(dispense_delay.duration)

    def mix(self, mix_properties: MixProperties, last_dispense_push_out: bool) -> None:
        """Execute mix steps.

        1. Use same flow rates and delays as aspirate and dispense
        2. Do [(aspirate + dispense) x repetitions] at the same position
        3. Do NOT push out at the end of dispense
        4. USE the delay property from aspirate & dispense during mix as well (flow rate and delay are coordinated with each other)
        5. Do not mix during consolidation
        NOTE: For most of our built-in definitions, we will keep _mix_ off because it is a very application specific thing.
        We should mention in our docs that users should adjust this property according to their application.
        """
        if not mix_properties.enabled:
            return
        # Assertion only for mypy purposes
        assert (
            mix_properties.repetitions is not None and mix_properties.volume is not None
        )
        push_out_vol = (
            self._transfer_properties.dispense.push_out_by_volume.get_for_volume(
                mix_properties.volume
            )
        )
        for n in range(mix_properties.repetitions, 0, -1):
            self.aspirate_and_wait(volume=mix_properties.volume)
            self.dispense_and_wait(
                dispense_properties=self._transfer_properties.dispense,  # TODO: check that using single-dispense props during mix is correct
                volume=mix_properties.volume,
                push_out_override=push_out_vol
                if last_dispense_push_out is True and n == 1
                else 0,
            )

    def pre_wet(
        self,
        volume: float,
    ) -> None:
        """Do a pre-wet.

        - 1 combo of aspirate + dispense at the same flow rate as specified in asp & disp and the delays in asp & disp
        - Use the target volume/ volume we will be aspirating
        - No push out
        - No pre-wet for consolidation
        """
        if not self._transfer_properties.aspirate.pre_wet:
            return
        mix_props = MixProperties(_enabled=True, _repetitions=1, _volume=volume)
        self.mix(mix_properties=mix_props, last_dispense_push_out=False)

    def retract_after_aspiration(
        self, volume: float, add_air_gap: Optional[bool] = True
    ) -> None:
        """Execute post-aspiration retraction steps.

        1. Move TO the position reference+offset AT the specified speed
            Raise error if retract is below aspirate position or below the meniscus
        2. Delay
        3. Touch tip
            - Move to the Z offset position
            - Touch tip to the sides at the specified speed (tip moves back to the center as part of touch tip)
            - Return back to the retract position
        4. Air gap
            - Air gap volume depends on the amount of liquid in the pipette
            So if total aspirated volume is 20, use the value for airGapByVolume[20]
            Flow rate = min(aspirateFlowRate, (airGapByVolume)/sec)
            - Use post-aspirate delay

        Args:
            volume: dispense volume
            add_air_gap: whether to add an air gap before moving away from the current well.
                         This value is True for all retractions, except when retracting
                         during a multi-dispense.
        """
        # TODO: Raise error if retract is below the meniscus
        retract_props = self._transfer_properties.aspirate.retract
        retract_point = absolute_point_from_position_reference_and_offset(
            well=self._target_well,
            position_reference=retract_props.position_reference,
            offset=retract_props.offset,
        )
        retract_location = Location(
            retract_point, labware=self._target_location.labware
        )
        tx_utils.raise_if_location_inside_liquid(
            location=retract_location,
            well_location=self._target_location,
            well_core=self._target_well,
            location_check_descriptors=LocationCheckDescriptors(
                location_type="retract end",
                pipetting_action="aspirate",
            ),
            logger=log,
        )
        self._instrument.move_to(
            location=retract_location,
            well_core=self._target_well,
            force_direct=True,
            minimum_z_height=None,
            speed=retract_props.speed,
        )
        retract_delay = retract_props.delay
        if retract_delay.enabled:
            assert retract_delay.duration is not None
            self._instrument.delay(retract_delay.duration)
        touch_tip_props = retract_props.touch_tip
        if touch_tip_props.enabled:
            assert (
                touch_tip_props.speed is not None
                and touch_tip_props.z_offset is not None
                and touch_tip_props.mm_to_edge is not None
            )
            self._instrument.touch_tip(
                location=retract_location,
                well_core=self._target_well,
                radius=1,
                z_offset=touch_tip_props.z_offset,
                speed=touch_tip_props.speed,
                mm_from_edge=touch_tip_props.mm_to_edge,
            )
            self._instrument.move_to(
                location=retract_location,
                well_core=self._target_well,
                force_direct=True,
                minimum_z_height=None,
                # Full speed because the tip will already be out of the liquid
                speed=None,
            )
        # For consolidate, we need to know the total amount that is in the pipette
        # since this may not be the first aspirate
        if self._transfer_type == TransferType.MANY_TO_ONE:
            volume_for_air_gap = self._instrument.get_current_volume()
        else:
            volume_for_air_gap = volume
        if add_air_gap:
            self._add_air_gap(
                air_gap_volume=self._transfer_properties.aspirate.retract.air_gap_by_volume.get_for_volume(
                    volume_for_air_gap
                )
            )

    def retract_after_dispensing(
        self,
        trash_location: Union[Location, TrashBin, WasteChute],
        source_location: Optional[Location],
        source_well: Optional[WellCore],
        add_final_air_gap: bool,
    ) -> None:
        """Execute post-dispense retraction steps.
        1. Position ref+offset is the ending position. Move to this position using specified speed
        2. If blowout is enabled and “destination”
            - Do blow-out (at the retract position)
            - Leave plunger down
        3. Touch-tip
        4. If not ready-to-aspirate
            - Prepare-to-aspirate (at the retract position)
        5. Air-gap (at the retract position)
            - This air gap is for preventing any stray droplets from falling while moving the pipette.
                It will be performed out of caution even if we just did a blow_out and should *hypothetically*
                have no liquid left in the tip.
            - This air gap will be removed at the next aspirate.
                If this is the last step of the transfer, and we aren't dropping the tip off,
                then the air gap will be left as is(?).
        6. If blowout is “source” or “trash”
            - Move to location (top of Well)
            - Do blow-out (top of well)
            - Do touch-tip (?????) (only if it’s in a non-trash location)
            - Prepare-to-aspirate (top of well)
            - Do air-gap (top of well)
        7. If drop tip, move to drop tip location, drop tip
        """
        # TODO: Raise error if retract is below the meniscus

        retract_props = self._transfer_properties.dispense.retract
        retract_point = absolute_point_from_position_reference_and_offset(
            well=self._target_well,
            position_reference=retract_props.position_reference,
            offset=retract_props.offset,
        )
        retract_location = Location(
            retract_point, labware=self._target_location.labware
        )
        tx_utils.raise_if_location_inside_liquid(
            location=retract_location,
            well_location=self._target_location,
            well_core=self._target_well,
            location_check_descriptors=LocationCheckDescriptors(
                location_type="retract end",
                pipetting_action="dispense",
            ),
            logger=log,
        )
        self._instrument.move_to(
            location=retract_location,
            well_core=self._target_well,
            force_direct=True,
            minimum_z_height=None,
            speed=retract_props.speed,
        )
        retract_delay = retract_props.delay
        if retract_delay.enabled:
            assert retract_delay.duration is not None
            self._instrument.delay(retract_delay.duration)

        blowout_props = retract_props.blowout
        if (
            blowout_props.enabled
            and blowout_props.location == BlowoutLocation.DESTINATION
        ):
            assert blowout_props.flow_rate is not None
            self._instrument.set_flow_rate(blow_out=blowout_props.flow_rate)
            self._instrument.blow_out(
                location=retract_location,
                well_core=None,
                in_place=True,
            )
            self._tip_state.ready_to_aspirate = False
        is_final_air_gap = (
            blowout_props.enabled
            and blowout_props.location == BlowoutLocation.DESTINATION
        ) or not blowout_props.enabled
        # Regardless of the blowout location, do touch tip and air gap
        # when leaving the dispense well. If this will be the final air gap, i.e,
        # we won't be moving to a Trash or a Source for Blowout after this air gap,
        # then skip the final air gap if we have been told to do so.
        self._do_touch_tip_and_air_gap(
            touch_tip_properties=retract_props.touch_tip,
            location=retract_location,
            well=self._target_well,
            add_air_gap=False if is_final_air_gap and not add_final_air_gap else True,
        )

        if (
            blowout_props.enabled
            and blowout_props.location != BlowoutLocation.DESTINATION
        ):
            # TODO: no-op touch tip if touch tip is enabled and blowout is in trash/ reservoir/ any labware with touch-tip disabled
            assert blowout_props.flow_rate is not None
            self._instrument.set_flow_rate(blow_out=blowout_props.flow_rate)
            touch_tip_and_air_gap_location: Optional[Location]
            if blowout_props.location == BlowoutLocation.SOURCE:
                if source_location is None or source_well is None:
                    raise RuntimeError(
                        "Blowout location is 'source' but source location &/or well is not provided."
                    )
                # TODO: check if we should add a blowout location z-offset in liq class definition
                self._instrument.blow_out(
                    location=Location(
                        source_well.get_top(0), labware=source_location.labware
                    ),
                    well_core=source_well,
                    in_place=False,
                )
                touch_tip_and_air_gap_location = Location(
                    source_well.get_top(0), labware=source_location.labware
                )
                touch_tip_and_air_gap_well = source_well
            else:
                self._instrument.blow_out(
                    location=trash_location,
                    well_core=None,
                    in_place=False,
                )
                touch_tip_and_air_gap_location = (
                    trash_location if isinstance(trash_location, Location) else None
                )
                touch_tip_and_air_gap_well = (
                    # We have already established that trash location of `Location` type
                    # has its `labware` as `Well` type.
                    trash_location.labware.as_well()._core  # type: ignore[assignment]
                    if isinstance(trash_location, Location)
                    else None
                )
            last_air_gap = self._tip_state.last_liquid_and_air_gap_in_tip.air_gap
            self._tip_state.delete_air_gap(last_air_gap)
            self._tip_state.ready_to_aspirate = False
            # Do touch tip and air gap again after blowing out into source well or trash
            self._do_touch_tip_and_air_gap(
                touch_tip_properties=retract_props.touch_tip,
                location=touch_tip_and_air_gap_location,
                well=touch_tip_and_air_gap_well,
                add_air_gap=add_final_air_gap,
            )

    def retract_during_multi_dispensing(
        self,
        trash_location: Union[Location, TrashBin, WasteChute],
        source_location: Optional[Location],
        source_well: Optional[WellCore],
        conditioning_volume: float,
        add_final_air_gap: bool,
        is_last_retract: bool,
    ) -> None:
        """Execute post-dispense retraction steps when the dispense is a part of a multi-dispense.

        Args:
            trash_location: Location where we can drop tips or blowout, if set to do so
            source_location: Location where we can blowout, if set to do so
            source_well: Well where we can blowout, if set to do so
            conditioning_volume: Conditioning volume used for this multi-dispense. Can be 0
            add_final_air_gap: Whether we should add the final air gap of the step
            is_last_retract: Whether this is the last retract of the multi-dispense steps,
                i.e., this is part of the last dispense in the series of consecutive dispenses.
                This dispense might not be the last dispense of the entire distribution.

        This function is mostly similar to the single-dispense retract function except
        that it handles air gaps differently based on the disposal volume, conditioning volume
        and whether we are moving to another dispense or going back to the source.
        """
        # TODO: Raise error if retract is below the meniscus

        assert self._transfer_properties.multi_dispense is not None

        retract_props = self._transfer_properties.multi_dispense.retract
        retract_point = absolute_point_from_position_reference_and_offset(
            well=self._target_well,
            position_reference=retract_props.position_reference,
            offset=retract_props.offset,
        )
        retract_location = Location(
            retract_point, labware=self._target_location.labware
        )
        tx_utils.raise_if_location_inside_liquid(
            location=retract_location,
            well_location=self._target_location,
            well_core=self._target_well,
            location_check_descriptors=LocationCheckDescriptors(
                location_type="retract end",
                pipetting_action="dispense",
            ),
            logger=log,
        )
        self._instrument.move_to(
            location=retract_location,
            well_core=self._target_well,
            force_direct=True,
            minimum_z_height=None,
            speed=retract_props.speed,
        )
        retract_delay = retract_props.delay
        if retract_delay.enabled:
            assert retract_delay.duration is not None
            self._instrument.delay(retract_delay.duration)

        blowout_props = retract_props.blowout
        if (
            is_last_retract
            and blowout_props.enabled
            and blowout_props.location == BlowoutLocation.DESTINATION
        ):
            assert blowout_props.flow_rate is not None
            self._instrument.set_flow_rate(blow_out=blowout_props.flow_rate)
            self._instrument.blow_out(
                location=retract_location,
                well_core=None,
                in_place=True,
            )
            self._tip_state.ready_to_aspirate = False

        # A retract will perform total of two air gaps if we need to blow out in source or trash:
        #   - 1st air gap: added before leaving the destination volume to go to src/ trash
        #   - 2nd air gap: added before leaving the blowout location to go to src or tip drop location
        # But if blowout is disabled or is set to Destination well, then only one air gap
        # will be added after retracting, before moving to src or tip drop location.
        # `is_final_air_gap_of_current_retract` tells us whether the next air gap
        # we will be adding, is going to be the last air gap of this step.
        is_final_air_gap_of_current_retract = (
            blowout_props.enabled
            and blowout_props.location == BlowoutLocation.DESTINATION
        ) or not blowout_props.enabled

        # Whether we should add the next air gap depends on the cases as shown below.
        # The main points when deciding this-
        #   - When we have used a conditioning volume, we do not want to add air gaps
        #     while there's still liquid in tip for dispensing
        #   - If we are not using conditioning volume then we want to add gaps just like
        #     we do during the one-to-one transfers
        #   - If this will be the last air gap of the step, if the above two conditions
        #     indicate that we should be adding an air gap, use `add_final_air_gap` as
        #     the final decider of whether to add the air gap.
        if is_final_air_gap_of_current_retract:
            if conditioning_volume > 0:
                add_air_gap = is_last_retract and add_final_air_gap
            else:
                add_air_gap = add_final_air_gap
        else:
            if conditioning_volume > 0:
                add_air_gap = is_last_retract
            else:
                add_air_gap = True

        # Regardless of the blowout location, do touch tip
        # when leaving the dispense well.
        # Add an air gap depending on conditioning volume + whether this is
        # the last step of a multi-dispense sequence + whether this is the last step
        # of the entire liquid distribution.
        self._do_touch_tip_and_air_gap(
            touch_tip_properties=retract_props.touch_tip,
            location=retract_location,
            well=self._target_well,
            add_air_gap=add_air_gap,
        )

        if (
            is_last_retract  # We can do a blowout only on the last multi-dispense step
            and blowout_props.enabled
            and blowout_props.location != BlowoutLocation.DESTINATION
        ):
            assert blowout_props.flow_rate is not None
            self._instrument.set_flow_rate(blow_out=blowout_props.flow_rate)
            touch_tip_and_air_gap_location: Optional[Location]
            if blowout_props.location == BlowoutLocation.SOURCE:
                if source_location is None or source_well is None:
                    raise RuntimeError(
                        "Blowout location is 'source' but source location &/or well is not provided."
                    )
                # TODO: check if we should add a blowout location z-offset in liq class definition
                self._instrument.blow_out(
                    location=Location(
                        source_well.get_top(0), labware=source_location.labware
                    ),
                    well_core=source_well,
                    in_place=False,
                )
                touch_tip_and_air_gap_location = Location(
                    source_well.get_top(0), labware=source_location.labware
                )
                touch_tip_and_air_gap_well = source_well
            else:
                self._instrument.blow_out(
                    location=trash_location,
                    well_core=None,
                    in_place=False,
                )
                touch_tip_and_air_gap_location = (
                    trash_location if isinstance(trash_location, Location) else None
                )
                touch_tip_and_air_gap_well = (
                    # We have already established that trash location of `Location` type
                    # has its `labware` as `Well` type.
                    trash_location.labware.as_well()._core  # type: ignore[assignment]
                    if isinstance(trash_location, Location)
                    else None
                )
            last_air_gap = self._tip_state.last_liquid_and_air_gap_in_tip.air_gap
            self._tip_state.delete_air_gap(last_air_gap)
            self._tip_state.ready_to_aspirate = False

            # Do touch tip and air gap again after blowing out into source well or trash
            self._do_touch_tip_and_air_gap(
                touch_tip_properties=retract_props.touch_tip,
                location=touch_tip_and_air_gap_location,
                well=touch_tip_and_air_gap_well,
                add_air_gap=(
                    # Same check as before for when it's the final air gap of current retract
                    conditioning_volume > 0
                    and is_last_retract
                    and add_final_air_gap
                ),
            )

    def _do_touch_tip_and_air_gap(
        self,
        touch_tip_properties: TouchTipProperties,
        location: Optional[Location],
        well: Optional[WellCore],
        add_air_gap: bool,
    ) -> None:
        """Perform touch tip and air gap as part of post-dispense retract."""
        if touch_tip_properties.enabled:
            assert (
                touch_tip_properties.speed is not None
                and touch_tip_properties.z_offset is not None
                and touch_tip_properties.mm_to_edge is not None
            )
            # TODO:, check that when blow out is a non-dest-well,
            #  whether the touch tip params from transfer props should be used for
            #  both dest-well touch tip and non-dest-well touch tip.
            if well is not None and location is not None:
                try:
                    self._instrument.touch_tip(
                        location=location,
                        well_core=well,
                        radius=1,
                        z_offset=touch_tip_properties.z_offset,
                        speed=touch_tip_properties.speed,
                        mm_from_edge=touch_tip_properties.mm_to_edge,
                    )
                except TouchTipDisabledError:
                    # TODO: log a warning
                    pass

                # Move back to the 'retract' position
                self._instrument.move_to(
                    location=location,
                    well_core=well,
                    force_direct=True,
                    minimum_z_height=None,
                    # Full speed because the tip will already be out of the liquid
                    speed=None,
                )

        # TODO: check if it is okay to just do `prepare_to_aspirate` unconditionally
        if not self._tip_state.ready_to_aspirate:
            self._instrument.prepare_to_aspirate()
            self._tip_state.ready_to_aspirate = True
        if add_air_gap:
            self._add_air_gap(
                air_gap_volume=self._transfer_properties.aspirate.retract.air_gap_by_volume.get_for_volume(
                    0
                )
            )

    def _add_air_gap(self, air_gap_volume: float) -> None:
        """Add an air gap."""
        if air_gap_volume == 0:
            return
        aspirate_props = self._transfer_properties.aspirate
        correction_volume = aspirate_props.correction_by_volume.get_for_volume(
            air_gap_volume
        )
        # The maximum flow rate should be air_gap_volume per second
        flow_rate = min(
            aspirate_props.flow_rate_by_volume.get_for_volume(air_gap_volume),
            air_gap_volume,
        )
        self._instrument.air_gap_in_place(
            volume=air_gap_volume,
            flow_rate=flow_rate,
            correction_volume=correction_volume,
        )
        delay_props = aspirate_props.delay
        if delay_props.enabled:
            # Assertion only for mypy purposes
            assert delay_props.duration is not None
            self._instrument.delay(delay_props.duration)
        self._tip_state.append_air_gap(air_gap_volume)

    def _remove_air_gap(self, location: Location) -> None:
        """Remove a previously added air gap."""
        last_air_gap = self._tip_state.last_liquid_and_air_gap_in_tip.air_gap
        if last_air_gap == 0:
            return

        dispense_props = self._transfer_properties.dispense
        correction_volume = dispense_props.correction_by_volume.get_for_volume(
            last_air_gap
        )
        # The maximum flow rate should be air_gap_volume per second
        flow_rate = min(
            dispense_props.flow_rate_by_volume.get_for_volume(last_air_gap),
            last_air_gap,
        )
        self._instrument.dispense(
            location=location,
            well_core=None,
            volume=last_air_gap,
            rate=1,
            flow_rate=flow_rate,
            in_place=True,
            push_out=0,
            correction_volume=correction_volume,
        )
        self._tip_state.delete_air_gap(last_air_gap)
        dispense_delay = dispense_props.delay
        if dispense_delay.enabled:
            assert dispense_delay.duration is not None
            self._instrument.delay(dispense_delay.duration)


def absolute_point_from_position_reference_and_offset(
    well: WellCore,
    position_reference: PositionReference,
    offset: Coordinate,
) -> Point:
    """Return the absolute point, given the well, the position reference and offset."""
    match position_reference:
        case PositionReference.WELL_TOP:
            reference_point = well.get_top(0)
        case PositionReference.WELL_BOTTOM:
            reference_point = well.get_bottom(0)
        case PositionReference.WELL_CENTER:
            reference_point = well.get_center()
        case PositionReference.LIQUID_MENISCUS:
            meniscus_point = well.get_meniscus()
            if not isinstance(meniscus_point, Point):
                reference_point = well.get_center()
            else:
                reference_point = meniscus_point
        case _:
            raise ValueError(f"Unknown position reference {position_reference}")
    return reference_point + Point(offset.x, offset.y, offset.z)



"""Wait for resume command request, result, and implementation models."""
from __future__ import annotations
from typing import TYPE_CHECKING, Optional, Type, Any

from pydantic import BaseModel, Field
from pydantic.json_schema import SkipJsonSchema
from typing_extensions import Literal

from .command import AbstractCommandImpl, BaseCommand, BaseCommandCreate, SuccessData
from ..errors.error_occurrence import ErrorOccurrence

if TYPE_CHECKING:
    from ..execution import RunControlHandler


# NOTE: multiple values accepted for backwards compatibility
# with the 6.0.0-beta.0 release, which used `pause`
WaitForResumeCommandType = Literal["waitForResume", "pause"]


def _remove_default(s: dict[str, Any]) -> None:
    s.pop("default", None)


class WaitForResumeParams(BaseModel):
    """Payload required to pause the protocol."""

    message: str | SkipJsonSchema[None] = Field(
        None,
        description="A user-facing message associated with the pause",
        json_schema_extra=_remove_default,
    )


class WaitForResumeResult(BaseModel):
    """Result data from the execution of a WaitForResume command."""


class WaitForResumeImplementation(
    AbstractCommandImpl[WaitForResumeParams, SuccessData[WaitForResumeResult]]
):
    """Wait for resume command implementation."""

    def __init__(self, run_control: RunControlHandler, **kwargs: object) -> None:
        self._run_control = run_control

    async def execute(
        self, params: WaitForResumeParams
    ) -> SuccessData[WaitForResumeResult]:
        """Dispatch a PauseAction to the store to pause the protocol."""
        await self._run_control.wait_for_resume()
        return SuccessData(
            public=WaitForResumeResult(),
        )


class WaitForResume(
    BaseCommand[WaitForResumeParams, WaitForResumeResult, ErrorOccurrence]
):
    """Wait for resume command model."""

    commandType: WaitForResumeCommandType = "waitForResume"
    params: WaitForResumeParams
    result: Optional[WaitForResumeResult] = None

    _ImplementationCls: Type[WaitForResumeImplementation] = WaitForResumeImplementation


class WaitForResumeCreate(BaseCommandCreate[WaitForResumeParams]):
    """Wait for resume command request model."""

    commandType: WaitForResumeCommandType = "waitForResume"
    params: WaitForResumeParams

    _CommandCls: Type[WaitForResume] = WaitForResume



"""Geometry state getters."""

from logging import getLogger
import enum
from numpy import array, dot, double as npdouble
from numpy.typing import NDArray
from typing import Optional, List, Tuple, Union, cast, TypeVar, Dict, Set
from dataclasses import dataclass
from functools import cached_property

from opentrons.types import (
    Point,
    DeckSlotName,
    StagingSlotName,
    MountType,
    MeniscusTrackingTarget,
)

from opentrons_shared_data.errors.exceptions import InvalidStoredData
from opentrons_shared_data.labware.constants import WELL_NAME_PATTERN
from opentrons_shared_data.labware.labware_definition import LabwareDefinition
from opentrons_shared_data.deck.types import CutoutFixture
from opentrons_shared_data.pipette import PIPETTE_X_SPAN
from opentrons_shared_data.pipette.types import ChannelCount, LabwareUri

from .. import errors
from ..errors import (
    LabwareNotLoadedError,
    LabwareNotLoadedOnLabwareError,
    LabwareNotLoadedOnModuleError,
    LabwareMovementNotAllowedError,
    OperationLocationNotInWellError,
)
from ..resources import (
    fixture_validation,
    labware_validation,
    deck_configuration_provider,
)
from ..types import (
    OFF_DECK_LOCATION,
    SYSTEM_LOCATION,
    LoadedLabware,
    LoadedModule,
    WellLocation,
    LiquidHandlingWellLocation,
    DropTipWellLocation,
    PickUpTipWellLocation,
    WellOrigin,
    DropTipWellOrigin,
    WellOffset,
    DeckSlotLocation,
    ModuleLocation,
    OnLabwareLocation,
    LabwareLocation,
    LabwareOffsetVector,
    ModuleOffsetVector,
    ModuleOffsetData,
    CurrentWell,
    CurrentPipetteLocation,
    TipGeometry,
    LabwareMovementOffsetData,
    InStackerHopperLocation,
    OnDeckLabwareLocation,
    AddressableAreaLocation,
    AddressableOffsetVector,
    StagingSlotLocation,
    LabwareOffsetLocationSequence,
    OnModuleOffsetLocationSequenceComponent,
    OnAddressableAreaOffsetLocationSequenceComponent,
    OnLabwareOffsetLocationSequenceComponent,
    OnLabwareLocationSequenceComponent,
    ModuleModel,
    PotentialCutoutFixture,
    LabwareLocationSequence,
    OnModuleLocationSequenceComponent,
    OnAddressableAreaLocationSequenceComponent,
    OnCutoutFixtureLocationSequenceComponent,
    NotOnDeckLocationSequenceComponent,
    AreaType,
    labware_location_is_off_deck,
    labware_location_is_system,
)
from ..types.liquid_level_detection import SimulatedProbeResult, LiquidTrackingType
from .config import Config
from .labware import LabwareView
from .wells import WellView
from .modules import ModuleView
from .pipettes import PipetteView
from .addressable_areas import AddressableAreaView
from .frustum_helpers import (
    find_volume_at_well_height,
    find_height_at_well_volume,
)
from ._well_math import wells_covered_by_pipette_configuration, nozzles_per_well


_LOG = getLogger(__name__)
SLOT_WIDTH = 128
_PIPETTE_HOMED_POSITION_Z = (
    248.0  # Height of the bottom of the nozzle without the tip attached when homed
)


class _TipDropSection(enum.Enum):
    """Well sections to drop tips in."""

    LEFT = "left"
    RIGHT = "right"


class _GripperMoveType(enum.Enum):
    """Types of gripper movement."""

    PICK_UP_LABWARE = enum.auto()
    DROP_LABWARE = enum.auto()


@dataclass
class _AbsoluteRobotExtents:
    front_left: Dict[MountType, Point]
    back_right: Dict[MountType, Point]
    deck_extents: Point
    padding_rear: float
    padding_front: float
    padding_left_side: float
    padding_right_side: float


_LabwareLocation = TypeVar("_LabwareLocation", bound=LabwareLocation)


# TODO(mc, 2021-06-03): continue evaluation of which selectors should go here
# vs which selectors should be in LabwareView
class GeometryView:
    """Geometry computed state getters."""

    def __init__(
        self,
        config: Config,
        labware_view: LabwareView,
        well_view: WellView,
        module_view: ModuleView,
        pipette_view: PipetteView,
        addressable_area_view: AddressableAreaView,
    ) -> None:
        """Initialize a GeometryView instance."""
        self._config = config
        self._labware = labware_view
        self._wells = well_view
        self._modules = module_view
        self._pipettes = pipette_view
        self._addressable_areas = addressable_area_view
        self._last_drop_tip_location_spot: Dict[str, _TipDropSection] = {}

    @cached_property
    def absolute_deck_extents(self) -> _AbsoluteRobotExtents:
        """The absolute deck extents for a given robot deck."""
        left_offset = self._addressable_areas.mount_offsets["left"]
        right_offset = self._addressable_areas.mount_offsets["right"]

        front_left_abs = {
            MountType.LEFT: Point(left_offset.x, -1 * left_offset.y, left_offset.z),
            MountType.RIGHT: Point(right_offset.x, -1 * right_offset.y, right_offset.z),
        }
        back_right_abs = {
            MountType.LEFT: self._addressable_areas.deck_extents + left_offset,
            MountType.RIGHT: self._addressable_areas.deck_extents + right_offset,
        }
        return _AbsoluteRobotExtents(
            front_left=front_left_abs,
            back_right=back_right_abs,
            deck_extents=self._addressable_areas.deck_extents,
            padding_rear=self._addressable_areas.padding_offsets["rear"],
            padding_front=self._addressable_areas.padding_offsets["front"],
            padding_left_side=self._addressable_areas.padding_offsets["left_side"],
            padding_right_side=self._addressable_areas.padding_offsets["right_side"],
        )

    def get_labware_highest_z(self, labware_id: str) -> float:
        """Get the highest Z-point of a labware."""
        labware_data = self._labware.get(labware_id)

        return self._get_highest_z_from_labware_data(labware_data)

    def get_all_obstacle_highest_z(self) -> float:
        """Get the highest Z-point across all obstacles that the instruments need to fly over."""
        highest_labware_z = max(
            (
                self._get_highest_z_from_labware_data(lw_data)
                for lw_data in self._labware.get_all()
                if lw_data.location != OFF_DECK_LOCATION
                and not self._labware.get_labware_by_lid_id(lw_data.id)
            ),
            default=0.0,
        )

        # Fixme (spp, 2023-12-04): the overall height is not the true highest z of modules
        #  on a Flex.
        highest_module_z = max(
            (
                self._modules.get_overall_height(module.id)
                for module in self._modules.get_all()
            ),
            default=0.0,
        )

        cutout_fixture_names = self._addressable_areas.get_all_cutout_fixtures()
        if cutout_fixture_names is None:
            # We're using a simulated deck config (see `Config.use_simulated_deck_config`).
            # We only know the addressable areas referenced by the protocol, not the fixtures
            # providing them. And there is more than one possible configuration of fixtures
            # to provide them. So, we can't know what the highest fixture is. Default to 0.
            #
            # Defaulting to 0 may not be the right thing to do here.
            # For example, suppose a protocol references an addressable area that implies a tall
            # fixture must be on the deck, and then it uses long tips that wouldn't be able to
            # clear the top of that fixture. We should perhaps raise an analysis error for that,
            # but defaulting to 0 here means we won't.
            highest_fixture_z = 0.0
        else:
            highest_fixture_z = max(
                (
                    self._addressable_areas.get_fixture_height(cutout_fixture_name)
                    for cutout_fixture_name in cutout_fixture_names
                ),
                default=0.0,
            )

        return max(
            highest_labware_z,
            highest_module_z,
            highest_fixture_z,
        )

    def get_highest_z_in_slot(
        self, slot: Union[DeckSlotLocation, StagingSlotLocation]
    ) -> float:
        """Get the highest Z-point of all items stacked in the given deck slot.

        This height includes the height of any module that occupies the given slot
        even if it wasn't loaded in that slot (e.g., thermocycler).
        """
        slot_item = self.get_slot_item(slot.slotName)
        if isinstance(slot_item, LoadedModule):
            # get height of module + all labware on it
            module_id = slot_item.id
            try:
                labware_id = self._labware.get_id_by_module(module_id=module_id)
            except LabwareNotLoadedOnModuleError:
                return self._modules.get_module_highest_z(
                    module_id=module_id,
                    addressable_areas=self._addressable_areas,
                )
            else:
                return self.get_highest_z_of_labware_stack(labware_id)
        elif isinstance(slot_item, LoadedLabware):
            # get stacked heights of all labware in the slot
            return self.get_highest_z_of_labware_stack(slot_item.id)
        elif type(slot_item) is dict:
            # TODO (cb, 2024-02-05): Eventually this logic should become the responsibility of bounding box
            # conflict checking, as fixtures may not always be considered as items from slots.
            return self._addressable_areas.get_fixture_height(slot_item["id"])
        else:
            return 0

    def get_highest_z_of_labware_stack(self, labware_id: str) -> float:
        """Get the highest Z-point of the topmost labware in the stack of labware on the given labware.

        If there is no labware on the given labware, returns highest z of the given labware.
        """
        try:
            stacked_labware_id = self._labware.get_id_by_labware(labware_id)
        except LabwareNotLoadedOnLabwareError:
            return self.get_labware_highest_z(labware_id)
        return self.get_highest_z_of_labware_stack(stacked_labware_id)

    def get_min_travel_z(
        self,
        pipette_id: str,
        labware_id: str,
        location: Optional[CurrentPipetteLocation],
        minimum_z_height: Optional[float],
    ) -> float:
        """Get the minimum allowed travel height of an arc move."""
        if (
            isinstance(location, CurrentWell)
            and pipette_id == location.pipette_id
            and labware_id == location.labware_id
        ):
            min_travel_z = self.get_labware_highest_z(labware_id)
        else:
            min_travel_z = self.get_all_obstacle_highest_z()
        if minimum_z_height:
            min_travel_z = max(min_travel_z, minimum_z_height)
        return min_travel_z

    def get_labware_parent_nominal_position(self, labware_id: str) -> Point:
        """Get the position of the labware's uncalibrated parent (deck slot, module, or another labware)."""
        try:
            addressable_area_name = self.get_ancestor_slot_name(labware_id).id
        except errors.LocationIsStagingSlotError:
            addressable_area_name = self._get_staging_slot_name(labware_id)
        except errors.LocationIsLidDockSlotError:
            addressable_area_name = self._get_lid_dock_slot_name(labware_id)
        parent_pos = self._addressable_areas.get_addressable_area_position(
            addressable_area_name
        )

        offset_from_parent = self._get_offset_from_parent(
            child_definition=self._labware.get_definition(labware_id),
            parent=self._labware.get(labware_id).location,
        )

        return Point(
            parent_pos.x + offset_from_parent.x,
            parent_pos.y + offset_from_parent.y,
            parent_pos.z + offset_from_parent.z,
        )

    def _get_offset_from_parent(
        self, child_definition: LabwareDefinition, parent: LabwareLocation
    ) -> LabwareOffsetVector:
        """Gets the offset vector of a labware placed on the given location.

        - For labware on Deck Slot: returns an offset of (0, 0, 0)
        - For labware on a Module: returns the nominal offset for the labware's position
          when placed on the specified module (using slot-transformed labwareOffset
          from the module's definition with any stacking overlap).
          Does not include module calibration offset or LPC offset.
        - For labware on another labware: returns the nominal offset for the labware
          as placed on the specified labware, taking into account any offsets for labware
          on modules as well as stacking overlaps.
          Does not include module calibration offset or LPC offset.
        """
        if isinstance(parent, (AddressableAreaLocation, DeckSlotLocation)):
            return LabwareOffsetVector(x=0, y=0, z=0)
        elif isinstance(parent, ModuleLocation):
            module_id = parent.moduleId
            module_model = self._modules.get_connected_model(module_id)
            stacking_overlap = self._labware.get_module_overlap_offsets(
                child_definition, module_model
            )
            module_to_child = self._modules.get_nominal_offset_to_child(
                module_id=module_id, addressable_areas=self._addressable_areas
            )
            return LabwareOffsetVector(
                x=module_to_child.x - stacking_overlap.x,
                y=module_to_child.y - stacking_overlap.y,
                z=module_to_child.z - stacking_overlap.z,
            )
        elif isinstance(parent, OnLabwareLocation):
            on_labware = self._labware.get(parent.labwareId)
            on_labware_dimensions = self._labware.get_dimensions(
                labware_id=on_labware.id
            )
            stacking_overlap = self._labware.get_labware_overlap_offsets(
                definition=child_definition, below_labware_name=on_labware.loadName
            )
            labware_offset = LabwareOffsetVector(
                x=stacking_overlap.x,
                y=stacking_overlap.y,
                z=on_labware_dimensions.z - stacking_overlap.z,
            )
            return labware_offset + self._get_offset_from_parent(
                self._labware.get_definition(on_labware.id), on_labware.location
            )
        else:
            raise errors.LabwareNotOnDeckError(
                "Cannot access labware since it is not on the deck. "
                "Either it has been loaded off-deck or its been moved off-deck."
            )

    def _normalize_module_calibration_offset(
        self,
        module_location: DeckSlotLocation,
        offset_data: Optional[ModuleOffsetData],
    ) -> ModuleOffsetVector:
        """Normalize the module calibration offset depending on the module location."""
        if not offset_data:
            return ModuleOffsetVector(x=0, y=0, z=0)
        offset = offset_data.moduleOffsetVector
        calibrated_slot = offset_data.location.slotName
        calibrated_slot_column = self.get_slot_column(calibrated_slot)
        current_slot_column = self.get_slot_column(module_location.slotName)
        # make sure that we have valid colums since we cant have modules in the middle of the deck
        assert set([calibrated_slot_column, current_slot_column]).issubset(
            {1, 3}
        ), f"Module calibration offset is an invalid slot {calibrated_slot}"

        # Check if the module has moved from one side of the deck to the other
        if calibrated_slot_column != current_slot_column:
            # Since the module was rotated, the calibration offset vector needs to be rotated by 180 degrees along the z axis
            saved_offset: NDArray[npdouble] = array([offset.x, offset.y, offset.z])
            rotation_matrix: NDArray[npdouble] = array(
                [[-1, 0, 0], [0, -1, 0], [0, 0, 1]]
            )
            new_offset = dot(saved_offset, rotation_matrix)
            offset = ModuleOffsetVector(
                x=new_offset[0], y=new_offset[1], z=new_offset[2]
            )
        return offset

    def _get_calibrated_module_offset(
        self, location: LabwareLocation
    ) -> ModuleOffsetVector:
        """Get a labware location's underlying calibrated module offset, if it is on a module."""
        if isinstance(location, ModuleLocation):
            module_id = location.moduleId
            module_location = self._modules.get_location(module_id)
            offset_data = self._modules.get_module_calibration_offset(module_id)
            return self._normalize_module_calibration_offset(
                module_location, offset_data
            )
        elif isinstance(location, (DeckSlotLocation, AddressableAreaLocation)):
            # TODO we might want to do a check here to make sure addressable area location is a standard deck slot
            #   and raise if its not (or maybe we don't actually care since modules will never be loaded elsewhere)
            return ModuleOffsetVector(x=0, y=0, z=0)
        elif isinstance(location, OnLabwareLocation):
            labware_data = self._labware.get(location.labwareId)
            return self._get_calibrated_module_offset(labware_data.location)
        elif (
            location == OFF_DECK_LOCATION
            or location == SYSTEM_LOCATION
            or isinstance(location, InStackerHopperLocation)
        ):
            raise errors.LabwareNotOnDeckError(
                "Labware does not have a slot or module associated with it"
                " since it is no longer on the deck."
            )

    def get_labware_parent_position(self, labware_id: str) -> Point:
        """Get the calibrated position of the labware's parent slot (deck or module)."""
        parent_pos = self.get_labware_parent_nominal_position(labware_id)
        labware_data = self._labware.get(labware_id)
        cal_offset = self._get_calibrated_module_offset(labware_data.location)

        return Point(
            x=parent_pos.x + cal_offset.x,
            y=parent_pos.y + cal_offset.y,
            z=parent_pos.z + cal_offset.z,
        )

    def get_labware_origin_position(self, labware_id: str) -> Point:
        """Get the position of the labware's origin, without calibration."""
        slot_pos = self.get_labware_parent_position(labware_id)
        origin_offset = self._labware.get_definition(labware_id).cornerOffsetFromSlot

        return Point(
            x=slot_pos.x + origin_offset.x,
            y=slot_pos.y + origin_offset.y,
            z=slot_pos.z + origin_offset.z,
        )

    def get_labware_position(self, labware_id: str) -> Point:
        """Get the calibrated origin of the labware."""
        origin_pos = self.get_labware_origin_position(labware_id)
        cal_offset = self._labware.get_labware_offset_vector(labware_id)

        return Point(
            x=origin_pos.x + cal_offset.x,
            y=origin_pos.y + cal_offset.y,
            z=origin_pos.z + cal_offset.z,
        )

    WellLocations = Union[
        WellLocation, LiquidHandlingWellLocation, PickUpTipWellLocation
    ]

    def validate_well_position(
        self,
        well_location: WellLocations,
        z_offset: float,
        pipette_id: Optional[str] = None,
    ) -> None:
        """Raise exception if operation location is not within well.

        Primarily this checks if there is not enough liquid in a well to do meniscus-relative static aspiration.
        """
        if well_location.origin == WellOrigin.MENISCUS:
            assert pipette_id is not None
            lld_min_height = self._pipettes.get_current_tip_lld_settings(
                pipette_id=pipette_id
            )
            if z_offset < lld_min_height:
                if isinstance(well_location, PickUpTipWellLocation):
                    raise OperationLocationNotInWellError(
                        f"Specifying {well_location.origin} with a height offset of {well_location.offset.z} results in a height of {z_offset} mm; the minimum allowed height for liquid tracking is {lld_min_height} mm"
                    )
                else:
                    raise OperationLocationNotInWellError(
                        f"Specifying {well_location.origin} with a height offset of {well_location.offset.z} results in a height of {z_offset} mm; the minimum allowed height for liquid tracking is {lld_min_height} mm"
                    )
        elif z_offset < 0:
            if isinstance(well_location, PickUpTipWellLocation):
                raise OperationLocationNotInWellError(
                    f"Specifying {well_location.origin} with an offset of {well_location.offset.z} results in a location below the bottom of the well"
                )
            else:
                raise OperationLocationNotInWellError(
                    f"Specifying {well_location.origin} with an offset of {well_location.offset.z} and a volume offset of {well_location.volumeOffset} results in a location below the bottom of the well"
                )

    def get_well_position(
        self,
        labware_id: str,
        well_name: str,
        well_location: Optional[WellLocations] = None,
        operation_volume: Optional[float] = None,
        pipette_id: Optional[str] = None,
    ) -> Point:
        """Given relative well location in a labware, get absolute position."""
        labware_pos = self.get_labware_position(labware_id)
        well_def = self._labware.get_well_definition(labware_id, well_name)
        well_depth = well_def.depth

        offset = WellOffset(x=0, y=0, z=well_depth)
        if well_location is not None:
            offset = well_location.offset
            offset_adjustment = self.get_well_offset_adjustment(
                labware_id=labware_id,
                well_name=well_name,
                well_location=well_location,
                well_depth=well_depth,
                operation_volume=operation_volume,
            )
            if not isinstance(offset_adjustment, SimulatedProbeResult):
                offset = offset.model_copy(update={"z": offset.z + offset_adjustment})
                self.validate_well_position(
                    well_location=well_location,
                    z_offset=offset.z,
                    pipette_id=pipette_id,
                )

        return Point(
            x=labware_pos.x + offset.x + well_def.x,
            y=labware_pos.y + offset.y + well_def.y,
            z=labware_pos.z + offset.z + well_def.z,
        )

    def get_nominal_well_position(
        self,
        labware_id: str,
        well_name: str,
    ) -> Point:
        """Get the well position without calibration offsets."""
        parent_pos = self.get_labware_parent_nominal_position(labware_id)
        origin_offset = self._labware.get_definition(labware_id).cornerOffsetFromSlot
        well_def = self._labware.get_well_definition(labware_id, well_name)
        return Point(
            x=parent_pos.x + origin_offset.x + well_def.x,
            y=parent_pos.y + origin_offset.y + well_def.y,
            z=parent_pos.z + origin_offset.z + well_def.z + well_def.depth,
        )

    def get_relative_well_location(
        self,
        labware_id: str,
        well_name: str,
        absolute_point: Point,
    ) -> WellLocation:
        """Given absolute position, get relative location of a well in a labware."""
        well_absolute_point = self.get_well_position(labware_id, well_name)
        delta = absolute_point - well_absolute_point

        return WellLocation(offset=WellOffset(x=delta.x, y=delta.y, z=delta.z))

    def get_relative_liquid_handling_well_location(
        self,
        labware_id: str,
        well_name: str,
        absolute_point: Point,
        meniscus_tracking: Optional[MeniscusTrackingTarget] = None,
    ) -> Tuple[LiquidHandlingWellLocation, bool]:
        """Given absolute position, get relative location of a well in a labware."""
        dynamic_liquid_tracking = False
        if meniscus_tracking:
            location = LiquidHandlingWellLocation(
                origin=WellOrigin.MENISCUS,
                offset=WellOffset(x=0, y=0, z=absolute_point.z),
            )
            if meniscus_tracking == MeniscusTrackingTarget.END:
                location.volumeOffset = "operationVolume"
            elif meniscus_tracking == MeniscusTrackingTarget.DYNAMIC:
                dynamic_liquid_tracking = True
        else:
            well_absolute_point = self.get_well_position(labware_id, well_name)
            delta = absolute_point - well_absolute_point
            location = LiquidHandlingWellLocation(
                offset=WellOffset(x=delta.x, y=delta.y, z=delta.z)
            )
        return location, dynamic_liquid_tracking

    def get_relative_pick_up_tip_well_location(
        self,
        labware_id: str,
        well_name: str,
        absolute_point: Point,
    ) -> PickUpTipWellLocation:
        """Given absolute position, get relative location of a well in a labware."""
        well_absolute_point = self.get_well_position(labware_id, well_name)
        delta = absolute_point - well_absolute_point

        return PickUpTipWellLocation(offset=WellOffset(x=delta.x, y=delta.y, z=delta.z))

    def get_well_height(
        self,
        labware_id: str,
        well_name: str,
    ) -> float:
        """Get the height of a specified well for a labware."""
        well_def = self._labware.get_well_definition(labware_id, well_name)
        return well_def.depth

    def _get_highest_z_from_labware_data(self, lw_data: LoadedLabware) -> float:
        labware_pos = self.get_labware_position(lw_data.id)
        definition = self._labware.get_definition(lw_data.id)
        z_dim = definition.dimensions.zDimension
        height_over_labware: float = 0
        if isinstance(lw_data.location, ModuleLocation):
            # Note: when calculating highest z of stacked labware, height-over-labware
            # gets accounted for only if the top labware is directly on the module.
            # So if there's a labware on an adapter on a module, then this
            # over-module-height gets ignored. We currently do not have any modules
            # that use an adapter and has height over labware so this doesn't cause
            # any issues yet. But if we add one in the future then this calculation
            # should be updated.
            module_id = lw_data.location.moduleId
            height_over_labware = self._modules.get_height_over_labware(module_id)
        return labware_pos.z + z_dim + height_over_labware

    def get_nominal_effective_tip_length(
        self,
        pipette_id: str,
        labware_id: str,
    ) -> float:
        """Given a labware and a pipette's config, get the nominal effective tip length.

        Effective tip length is the nominal tip length less the distance the
        tip overlaps with the pipette nozzle. This does not take calibrated
        tip lengths into account.
        """
        labware_uri = self._labware.get_definition_uri(labware_id)
        nominal_overlap = self._pipettes.get_nominal_tip_overlap(
            pipette_id=pipette_id, labware_uri=labware_uri
        )

        return self._labware.get_tip_length(
            labware_id=labware_id, overlap=nominal_overlap
        )

    def get_nominal_tip_geometry(
        self,
        pipette_id: str,
        labware_id: str,
        well_name: Optional[str],
    ) -> TipGeometry:
        """Given a labware, well, and hardware pipette config, get the tip geometry.

        Tip geometry includes effective tip length, tip diameter, and tip volume,
        which is all data required by the hardware controller for proper tip handling.

        This geometry data is based solely on labware and pipette definitions and
        does not take calibrated tip lengths into account.
        """
        effective_length = self.get_nominal_effective_tip_length(
            pipette_id=pipette_id,
            labware_id=labware_id,
        )
        well_def = self._labware.get_well_definition(labware_id, well_name)

        if well_def.shape != "circular":
            raise errors.LabwareIsNotTipRackError(
                f"Well {well_name} in labware {labware_id} is not circular."
            )

        return TipGeometry(
            length=effective_length,
            diameter=well_def.diameter,
            # TODO(mc, 2020-11-12): WellDefinition type says totalLiquidVolume
            #  is a float, but hardware controller expects an int
            volume=int(well_def.totalLiquidVolume),
        )

    def get_checked_tip_drop_location(
        self,
        pipette_id: str,
        labware_id: str,
        well_location: DropTipWellLocation,
        partially_configured: bool = False,
    ) -> WellLocation:
        """Get tip drop location given labware and hardware pipette.

        This makes sure that the well location has an appropriate origin & offset
        if one is not already set previously.
        """
        if (
            self._labware.get_definition(labware_id).parameters.isTiprack
            and partially_configured
        ):
            raise errors.UnexpectedProtocolError(
                "Cannot return tip to a tiprack while the pipette is configured for partial tip."
            )
        if well_location.origin != DropTipWellOrigin.DEFAULT:
            return WellLocation(
                origin=WellOrigin(well_location.origin.value),
                offset=well_location.offset,
            )

        if self._labware.get_definition(labware_id).parameters.isTiprack:
            z_offset = self._labware.get_tip_drop_z_offset(
                labware_id=labware_id,
                length_scale=self._pipettes.get_return_tip_scale(pipette_id),
                additional_offset=well_location.offset.z,
            )
        else:
            # return to top if labware is not tip rack
            z_offset = well_location.offset.z

        return WellLocation(
            origin=WellOrigin.TOP,
            offset=WellOffset(
                x=well_location.offset.x,
                y=well_location.offset.y,
                z=z_offset,
            ),
        )

    def convert_pick_up_tip_well_location(
        self, well_location: PickUpTipWellLocation
    ) -> WellLocation:
        """Convert PickUpTipWellLocation to WellLocation."""
        return WellLocation(
            origin=WellOrigin(well_location.origin.value), offset=well_location.offset
        )

    # TODO(jbl 11-30-2023) fold this function into get_ancestor_slot_name see RSS-411
    def _get_staging_slot_name(self, labware_id: str) -> str:
        """Get the staging slot name that the labware is on."""
        labware_location = self._labware.get(labware_id).location
        if isinstance(labware_location, OnLabwareLocation):
            below_labware_id = labware_location.labwareId
            return self._get_staging_slot_name(below_labware_id)
        elif isinstance(
            labware_location, AddressableAreaLocation
        ) and fixture_validation.is_staging_slot(labware_location.addressableAreaName):
            return labware_location.addressableAreaName
        else:
            raise ValueError(
                "Cannot get staging slot name for labware not on staging slot."
            )

    def _get_lid_dock_slot_name(self, labware_id: str) -> str:
        """Get the staging slot name that the labware is on."""
        labware_location = self._labware.get(labware_id).location
        assert isinstance(labware_location, AddressableAreaLocation)
        return labware_location.addressableAreaName

    def get_ancestor_slot_name(
        self, labware_id: str
    ) -> Union[DeckSlotName, StagingSlotName]:
        """Get the slot name of the labware or the module that the labware is on."""
        labware = self._labware.get(labware_id)
        slot_name: Union[DeckSlotName, StagingSlotName]

        if isinstance(labware.location, DeckSlotLocation):
            slot_name = labware.location.slotName
        elif isinstance(labware.location, ModuleLocation):
            module_id = labware.location.moduleId
            slot_name = self._modules.get_location(module_id).slotName
        elif isinstance(labware.location, OnLabwareLocation):
            below_labware_id = labware.location.labwareId
            slot_name = self.get_ancestor_slot_name(below_labware_id)
        elif isinstance(labware.location, AddressableAreaLocation):
            area_name = labware.location.addressableAreaName
            if self._labware.is_absorbance_reader_lid(labware_id):
                raise errors.LocationIsLidDockSlotError(
                    "Cannot get ancestor slot name for labware on lid dock slot."
                )
            elif fixture_validation.is_staging_slot(area_name):
                slot_name = StagingSlotName.from_primitive(area_name)
            else:
                slot_name = DeckSlotName.from_primitive(area_name)
        elif labware.location == OFF_DECK_LOCATION:
            raise errors.LabwareNotOnDeckError(
                f"Labware {labware_id} does not have a slot associated with it"
                f" since it is no longer on the deck."
            )

        return slot_name

    def ensure_location_not_occupied(
        self,
        location: _LabwareLocation,
        desired_addressable_area: Optional[str] = None,
    ) -> _LabwareLocation:
        """Ensure that the location does not already have either Labware or a Module in it."""
        # Collect set of existing fixtures, if any
        existing_fixtures = self._get_potential_fixtures_for_location_occupation(
            location
        )
        potential_fixtures = (
            self._get_potential_fixtures_for_location_occupation(
                AddressableAreaLocation(addressableAreaName=desired_addressable_area)
            )
            if desired_addressable_area is not None
            else None
        )

        # Handle the checking conflict on an incoming fixture
        if potential_fixtures is not None and isinstance(location, DeckSlotLocation):
            if (
                existing_fixtures is not None
                and not any(
                    location.slotName.id in fixture.provided_addressable_areas
                    for fixture in potential_fixtures[1].intersection(
                        existing_fixtures[1]
                    )
                )
            ) or (
                self._labware.get_by_slot(location.slotName) is not None
                and not any(
                    location.slotName.id in fixture.provided_addressable_areas
                    for fixture in potential_fixtures[1]
                )
            ):
                self._labware.raise_if_labware_in_location(location)

            else:
                self._modules.raise_if_module_in_location(location)

        # Otherwise handle standard conflict checking
        else:
            if isinstance(
                location,
                (
                    DeckSlotLocation,
                    ModuleLocation,
                    OnLabwareLocation,
                    AddressableAreaLocation,
                ),
            ):
                self._labware.raise_if_labware_in_location(location)

            area = (
                location.slotName.id
                if isinstance(location, DeckSlotLocation)
                else (
                    location.addressableAreaName
                    if isinstance(location, AddressableAreaLocation)
                    else None
                )
            )
            if area is not None and (
                existing_fixtures is None
                or not any(
                    area in fixture.provided_addressable_areas
                    for fixture in existing_fixtures[1]
                )
            ):
                if isinstance(location, DeckSlotLocation):
                    self._modules.raise_if_module_in_location(location)
                elif isinstance(location, AddressableAreaLocation):
                    self._modules.raise_if_module_in_location(
                        DeckSlotLocation(
                            slotName=self._addressable_areas.get_addressable_area_base_slot(
                                location.addressableAreaName
                            )
                        )
                    )

        return location

    def _get_potential_fixtures_for_location_occupation(
        self, location: _LabwareLocation
    ) -> Tuple[str, Set[PotentialCutoutFixture]] | None:
        loc: DeckSlotLocation | AddressableAreaLocation | None = None
        if isinstance(location, AddressableAreaLocation):
            # Convert the addressable area into a staging slot if applicable
            slots = StagingSlotName._value2member_map_
            for slot in slots:
                if location.addressableAreaName == slot:
                    loc = DeckSlotLocation(
                        slotName=DeckSlotName(location.addressableAreaName[0] + "3")
                    )
            if loc is None:
                loc = location
        elif isinstance(location, DeckSlotLocation):
            loc = location

        if isinstance(loc, DeckSlotLocation):
            module = self._modules.get_by_slot(loc.slotName)
            if module is not None and self._config.robot_type != "OT-2 Standard":
                fixtures = deck_configuration_provider.get_potential_cutout_fixtures(
                    addressable_area_name=self._modules.ensure_and_convert_module_fixture_location(
                        deck_slot=loc.slotName,
                        model=module.model,
                    ),
                    deck_definition=self._addressable_areas.deck_definition,
                )
            else:
                fixtures = None
        elif isinstance(loc, AddressableAreaLocation):
            fixtures = deck_configuration_provider.get_potential_cutout_fixtures(
                addressable_area_name=loc.addressableAreaName,
                deck_definition=self._addressable_areas.deck_definition,
            )
        else:
            fixtures = None
        return fixtures

    def _get_potential_disposal_location_cutout_fixtures(
        self, slot_name: DeckSlotName
    ) -> CutoutFixture | None:
        for area in self._addressable_areas.get_all():
            if (
                self._addressable_areas.get_addressable_area(area).area_type
                == AreaType.WASTE_CHUTE
                or self._addressable_areas.get_addressable_area(area).area_type
                == AreaType.MOVABLE_TRASH
            ) and slot_name == self._addressable_areas.get_addressable_area_base_slot(
                area
            ):
                # Given we only have one Waste Chute fixture and one type of Trash bin fixture it's
                # fine to return the first result of our potential fixtures here. This will need to
                # change in the future if there multiple trash fixtures that share the same area type.
                potential_fixture = (
                    deck_configuration_provider.get_potential_cutout_fixtures(
                        area, self._addressable_areas.deck_definition
                    )[1].pop()
                )
                return deck_configuration_provider.get_cutout_fixture(
                    potential_fixture.cutout_fixture_id,
                    self._addressable_areas.deck_definition,
                )
        return None

    def get_labware_grip_point(
        self,
        labware_definition: LabwareDefinition,
        location: Union[
            DeckSlotLocation, ModuleLocation, OnLabwareLocation, AddressableAreaLocation
        ],
    ) -> Point:
        """Get the grip point of the labware as placed on the given location.

        Returns the absolute position of the labware's gripping point as if
        it were placed on the specified location. Labware offset (LPC offset) not included.

        Grip point is the location where critical point of the gripper should move to
        in order to pick/drop the given labware in the specified location.
        It is calculated as the xy center of the slot with z as the point indicated by
        z-position of labware bottom + grip height from labware bottom.
        """
        grip_height_from_labware_bottom = (
            self._labware.get_grip_height_from_labware_bottom(labware_definition)
        )
        location_name: str
        module_location: ModuleLocation | None = None

        if isinstance(location, DeckSlotLocation):
            location_name = location.slotName.id
            offset = LabwareOffsetVector(x=0, y=0, z=0)
        elif isinstance(location, AddressableAreaLocation):
            location_name = location.addressableAreaName
            if fixture_validation.is_gripper_waste_chute(location_name):
                drop_labware_location = (
                    self._addressable_areas.get_addressable_area_move_to_location(
                        location_name
                    )
                )
                return drop_labware_location + Point(z=grip_height_from_labware_bottom)
            # Location should have been pre-validated so this will be a deck/staging area slot
            else:
                offset = LabwareOffsetVector(x=0, y=0, z=0)
        else:
            if isinstance(location, ModuleLocation):
                location_name = self._modules.get_provided_addressable_area(
                    location.moduleId
                )
                module_location = location
            else:  # OnLabwareLocation
                labware_loc = self._labware.get(location.labwareId).location
                if isinstance(labware_loc, ModuleLocation):
                    location_name = self._modules.get_provided_addressable_area(
                        labware_loc.moduleId
                    )
                    module_location = labware_loc
                else:
                    location_name = self.get_ancestor_slot_name(location.labwareId).id
            labware_offset = self._get_offset_from_parent(
                child_definition=labware_definition, parent=location
            )
            # Get the calibrated offset if the on labware location is on top of a module, otherwise return empty one
            cal_offset = self._get_calibrated_module_offset(location)
            offset = LabwareOffsetVector(
                x=labware_offset.x + cal_offset.x,
                y=labware_offset.y + cal_offset.y,
                z=labware_offset.z + cal_offset.z,
            )

        if module_location is not None:
            # Location center must be determined from the cutout the Module is loaded in
            position = deck_configuration_provider.get_cutout_position(
                cutout_id=self._addressable_areas.get_cutout_id_by_deck_slot_name(
                    self._modules.get_location(module_location.moduleId).slotName
                ),
                deck_definition=self._addressable_areas.deck_definition,
            )
            bounding_box = self._addressable_areas.get_addressable_area(
                location_name
            ).bounding_box
            location_center = Point(
                position.x + bounding_box.x / 2,
                position.y + bounding_box.y / 2,
                position.z,
            )

        else:
            location_center = self._addressable_areas.get_addressable_area_center(
                location_name
            )

        return Point(
            location_center.x + offset.x,
            location_center.y + offset.y,
            location_center.z + offset.z + grip_height_from_labware_bottom,
        )

    def get_extra_waypoints(
        self,
        location: Optional[CurrentPipetteLocation],
        to_slot: Union[DeckSlotName, StagingSlotName],
    ) -> List[Tuple[float, float]]:
        """Get extra waypoints for movement if thermocycler needs to be dodged."""
        if location is not None:
            if isinstance(location, CurrentWell):
                from_slot = self.get_ancestor_slot_name(location.labware_id)
            else:
                from_slot = self._addressable_areas.get_addressable_area_base_slot(
                    location.addressable_area_name
                )
            if self._modules.should_dodge_thermocycler(
                from_slot=from_slot, to_slot=to_slot
            ):
                middle_slot = DeckSlotName.SLOT_5.to_equivalent_for_robot_type(
                    self._config.robot_type
                )
                middle_slot_center = (
                    self._addressable_areas.get_addressable_area_center(
                        addressable_area_name=middle_slot.id,
                    )
                )
                return [(middle_slot_center.x, middle_slot_center.y)]
        return []

    def get_slot_item(
        self, slot_name: Union[DeckSlotName, StagingSlotName]
    ) -> Union[LoadedLabware, LoadedModule, CutoutFixture, None]:
        """Get the top-most item present in a deck slot, if any.

        This includes any module that occupies the given slot even if it wasn't loaded
        in that slot (e.g., thermocycler).
        """
        maybe_labware = self._labware.get_by_slot(
            slot_name=slot_name,
        )

        if isinstance(slot_name, DeckSlotName):
            maybe_fixture = self._addressable_areas.get_fixture_by_deck_slot_name(
                slot_name
            )

            # Ignore generic single slot fixtures
            if maybe_fixture and maybe_fixture["id"] in {
                "singleLeftSlot",
                "singleCenterSlot",
                "singleRightSlot",
            }:
                maybe_fixture = None

            maybe_module = self._modules.get_by_slot(
                slot_name=slot_name,
            ) or self._modules.get_overflowed_module_in_slot(slot_name=slot_name)

            # For situations in which the deck config is none
            if maybe_fixture is None and maybe_labware is None and maybe_module is None:
                # todo(chb 2025-03-19): This can go away once we solve the problem of no deck config in analysis
                maybe_fixture = self._get_potential_disposal_location_cutout_fixtures(
                    slot_name
                )
        else:
            # Modules and fixtures can't be loaded on staging slots
            maybe_fixture = None
            maybe_module = None

        return maybe_labware or maybe_module or maybe_fixture or None

    @staticmethod
    def get_slot_column(slot_name: Union[DeckSlotName, StagingSlotName]) -> int:
        """Get the column number for the specified slot."""
        if isinstance(slot_name, StagingSlotName):
            return 4
        row_col_name = slot_name.to_ot3_equivalent()
        slot_name_match = WELL_NAME_PATTERN.match(row_col_name.value)
        assert (
            slot_name_match is not None
        ), f"Slot name {row_col_name} did not match required pattern; please check labware location."

        row_name, column_name = slot_name_match.group(1, 2)
        return int(column_name)

    def get_next_tip_drop_location(
        self, labware_id: str, well_name: str, pipette_id: str
    ) -> DropTipWellLocation:
        """Get the next location within the specified well to drop the tip into.

        In order to prevent tip stacking, we will alternate between two tip drop locations:
        1. location in left section: a safe distance from left edge of the well
        2. location in right section: a safe distance from right edge of the well

        This safe distance for most cases would be a location where all tips drop
        reliably inside the labware's well. This can be calculated based off of the
        span of a pipette, including all its tips, in the x-direction.

        But we also need to account for the not-so-uncommon case of a left pipette
        trying to drop tips in a labware in the rightmost deck column and vice versa.
        If this labware extends beyond a regular deck slot, like the Flex's default trash,
        then even after keeping a margin for x-span of a pipette, we will get
        a location that's unreachable for the pipette. In such cases, we try to drop tips
        at the rightmost location that a left pipette is able to reach,
        and leftmost location that a right pipette is able to reach respectively.

        In these calculations we assume that the critical point of a pipette
        is considered to be the midpoint of the pipette's tip for single channel,
        and the midpoint of the entire tip assembly for multi-channel pipettes.
        We also assume that the pipette_x_span includes any safety margins required.
        """
        if not self._labware.is_fixed_trash(labware_id=labware_id):
            # In order to avoid the complexity of finding tip drop locations for
            # variety of labware with different well configs, we will allow
            # location cycling only for fixed trash labware right now.
            # TODO (spp, 2023-09-12): update this to possibly a labware-width based check,
            #  or a 'trash' quirk check, once movable trash is implemented.
            return DropTipWellLocation(
                origin=DropTipWellOrigin.DEFAULT,
                offset=WellOffset(x=0, y=0, z=0),
            )

        well_x_dim = self._labware.get_well_size(
            labware_id=labware_id, well_name=well_name
        )[0]
        pipette_channels = self._pipettes.get_config(pipette_id).channels
        pipette_mount = self._pipettes.get_mount(pipette_id)

        labware_slot_column = self.get_slot_column(
            slot_name=self.get_ancestor_slot_name(labware_id)
        )

        if self._last_drop_tip_location_spot.get(labware_id) == _TipDropSection.RIGHT:
            # Drop tip in LEFT section
            x_offset = self._get_drop_tip_well_x_offset(
                tip_drop_section=_TipDropSection.LEFT,
                well_x_dim=well_x_dim,
                pipette_channels=pipette_channels,
                pipette_mount=pipette_mount,
                labware_slot_column=labware_slot_column,
            )
            self._last_drop_tip_location_spot[labware_id] = _TipDropSection.LEFT
        else:
            # Drop tip in RIGHT section
            x_offset = self._get_drop_tip_well_x_offset(
                tip_drop_section=_TipDropSection.RIGHT,
                well_x_dim=well_x_dim,
                pipette_channels=pipette_channels,
                pipette_mount=pipette_mount,
                labware_slot_column=labware_slot_column,
            )
            self._last_drop_tip_location_spot[labware_id] = _TipDropSection.RIGHT

        return DropTipWellLocation(
            origin=DropTipWellOrigin.TOP,
            offset=WellOffset(
                x=x_offset,
                y=0,
                z=0,
            ),
        )

    # TODO find way to combine this with above
    def get_next_tip_drop_location_for_addressable_area(
        self,
        addressable_area_name: str,
        pipette_id: str,
    ) -> AddressableOffsetVector:
        """Get the next location within the specified well to drop the tip into.

        See the doc-string for `get_next_tip_drop_location` for more info on execution.
        """
        area_x_dim = self._addressable_areas.get_addressable_area(
            addressable_area_name
        ).bounding_box.x

        pipette_channels = self._pipettes.get_config(pipette_id).channels
        pipette_mount = self._pipettes.get_mount(pipette_id)

        labware_slot_column = self.get_slot_column(
            slot_name=self._addressable_areas.get_addressable_area_base_slot(
                addressable_area_name
            )
        )

        if (
            self._last_drop_tip_location_spot.get(addressable_area_name)
            == _TipDropSection.RIGHT
        ):
            # Drop tip in LEFT section
            x_offset = self._get_drop_tip_well_x_offset(
                tip_drop_section=_TipDropSection.LEFT,
                well_x_dim=area_x_dim,
                pipette_channels=pipette_channels,
                pipette_mount=pipette_mount,
                labware_slot_column=labware_slot_column,
            )
            self._last_drop_tip_location_spot[
                addressable_area_name
            ] = _TipDropSection.LEFT
        else:
            # Drop tip in RIGHT section
            x_offset = self._get_drop_tip_well_x_offset(
                tip_drop_section=_TipDropSection.RIGHT,
                well_x_dim=area_x_dim,
                pipette_channels=pipette_channels,
                pipette_mount=pipette_mount,
                labware_slot_column=labware_slot_column,
            )
            self._last_drop_tip_location_spot[
                addressable_area_name
            ] = _TipDropSection.RIGHT

        return AddressableOffsetVector(x=x_offset, y=0, z=0)

    @staticmethod
    def _get_drop_tip_well_x_offset(
        tip_drop_section: _TipDropSection,
        well_x_dim: float,
        pipette_channels: int,
        pipette_mount: MountType,
        labware_slot_column: int,
    ) -> float:
        """Get the well x offset for DropTipWellLocation."""
        drop_location_margin_from_labware_edge = (
            PIPETTE_X_SPAN[cast(ChannelCount, pipette_channels)] / 2
        )
        if tip_drop_section == _TipDropSection.LEFT:
            if (
                well_x_dim > SLOT_WIDTH
                and pipette_channels != 96
                and pipette_mount == MountType.RIGHT
                and labware_slot_column == 1
            ):
                # Pipette might not reach the default left spot so use a different left spot
                x_well_offset = (
                    -well_x_dim / 2 + drop_location_margin_from_labware_edge * 2
                )
            else:
                x_well_offset = -well_x_dim / 2 + drop_location_margin_from_labware_edge
                if x_well_offset > 0:
                    x_well_offset = 0
        else:
            if (
                well_x_dim > SLOT_WIDTH
                and pipette_channels != 96
                and pipette_mount == MountType.LEFT
                and labware_slot_column == 3
            ):
                # Pipette might not reach the default right spot so use a different right spot
                x_well_offset = (
                    -well_x_dim / 2
                    + SLOT_WIDTH
                    - drop_location_margin_from_labware_edge
                )
            else:
                x_well_offset = well_x_dim / 2 - drop_location_margin_from_labware_edge
                if x_well_offset < 0:
                    x_well_offset = 0
        return x_well_offset

    def get_final_labware_movement_offset_vectors(
        self,
        from_location: OnDeckLabwareLocation,
        to_location: OnDeckLabwareLocation,
        additional_offset_vector: LabwareMovementOffsetData,
        current_labware: LabwareDefinition,
    ) -> LabwareMovementOffsetData:
        """Calculate the final labware offset vector to use in labware movement."""
        pick_up_offset = (
            self.get_total_nominal_gripper_offset_for_move_type(
                location=from_location,
                move_type=_GripperMoveType.PICK_UP_LABWARE,
                current_labware=current_labware,
            )
            + additional_offset_vector.pickUpOffset
        )
        drop_offset = (
            self.get_total_nominal_gripper_offset_for_move_type(
                location=to_location,
                move_type=_GripperMoveType.DROP_LABWARE,
                current_labware=current_labware,
            )
            + additional_offset_vector.dropOffset
        )

        return LabwareMovementOffsetData(
            pickUpOffset=pick_up_offset, dropOffset=drop_offset
        )

    @staticmethod
    def ensure_valid_gripper_location(
        location: LabwareLocation,
    ) -> Union[
        DeckSlotLocation, ModuleLocation, OnLabwareLocation, AddressableAreaLocation
    ]:
        """Ensure valid on-deck location for gripper, otherwise raise error."""
        if not isinstance(
            location,
            (
                DeckSlotLocation,
                ModuleLocation,
                OnLabwareLocation,
                AddressableAreaLocation,
            ),
        ):
            raise errors.LabwareMovementNotAllowedError(
                "Off-deck labware movements are not supported using the gripper."
            )
        return location

    def get_total_nominal_gripper_offset_for_move_type(
        self,
        location: OnDeckLabwareLocation,
        move_type: _GripperMoveType,
        current_labware: LabwareDefinition,
    ) -> LabwareOffsetVector:
        """Get the total of the offsets to be used to pick up labware in its current location."""
        if move_type == _GripperMoveType.PICK_UP_LABWARE:
            if isinstance(
                location, (ModuleLocation, DeckSlotLocation, AddressableAreaLocation)
            ):
                return self._nominal_gripper_offsets_for_location(location).pickUpOffset
            else:
                # If it's a labware on a labware (most likely an adapter),
                # we calculate the offset as sum of offsets for the direct parent labware
                # and the underlying non-labware parent location.
                direct_parent_offset = self._nominal_gripper_offsets_for_location(
                    location
                )
                ancestor = self._labware.get_parent_location(location.labwareId)
                extra_offset = LabwareOffsetVector(x=0, y=0, z=0)
                if (
                    isinstance(ancestor, ModuleLocation)
                    and self._modules._state.requested_model_by_id[ancestor.moduleId]
                    == ModuleModel.THERMOCYCLER_MODULE_V2
                    and labware_validation.validate_definition_is_lid(current_labware)
                ):
                    if "lidOffsets" in current_labware.gripperOffsets.keys():
                        extra_offset = LabwareOffsetVector(
                            x=current_labware.gripperOffsets[
                                "lidOffsets"
                            ].pickUpOffset.x,
                            y=current_labware.gripperOffsets[
                                "lidOffsets"
                            ].pickUpOffset.y,
                            z=current_labware.gripperOffsets[
                                "lidOffsets"
                            ].pickUpOffset.z,
                        )
                    else:
                        raise errors.LabwareOffsetDoesNotExistError(
                            f"Labware Definition {current_labware.parameters.loadName} does not contain required field 'lidOffsets' of 'gripperOffsets'."
                        )

                assert isinstance(
                    ancestor,
                    (
                        DeckSlotLocation,
                        ModuleLocation,
                        OnLabwareLocation,
                        AddressableAreaLocation,
                    ),
                ), "No gripper offsets for off-deck labware"
                return (
                    direct_parent_offset.pickUpOffset
                    + self._nominal_gripper_offsets_for_location(
                        location=ancestor
                    ).pickUpOffset
                    + extra_offset
                )
        else:
            if isinstance(
                location, (ModuleLocation, DeckSlotLocation, AddressableAreaLocation)
            ):
                return self._nominal_gripper_offsets_for_location(location).dropOffset
            else:
                # If it's a labware on a labware (most likely an adapter),
                # we calculate the offset as sum of offsets for the direct parent labware
                # and the underlying non-labware parent location.
                direct_parent_offset = self._nominal_gripper_offsets_for_location(
                    location
                )
                ancestor = self._labware.get_parent_location(location.labwareId)
                extra_offset = LabwareOffsetVector(x=0, y=0, z=0)
                if (
                    isinstance(ancestor, ModuleLocation)
                    # todo(mm, 2024-11-06): Do not access private module state; only use public ModuleView methods.
                    and self._modules._state.requested_model_by_id[ancestor.moduleId]
                    == ModuleModel.THERMOCYCLER_MODULE_V2
                    and labware_validation.validate_definition_is_lid(current_labware)
                ):
                    if "lidOffsets" in current_labware.gripperOffsets.keys():
                        extra_offset = LabwareOffsetVector(
                            x=current_labware.gripperOffsets[
                                "lidOffsets"
                            ].pickUpOffset.x,
                            y=current_labware.gripperOffsets[
                                "lidOffsets"
                            ].pickUpOffset.y,
                            z=current_labware.gripperOffsets[
                                "lidOffsets"
                            ].pickUpOffset.z,
                        )
                    else:
                        raise errors.LabwareOffsetDoesNotExistError(
                            f"Labware Definition {current_labware.parameters.loadName} does not contain required field 'lidOffsets' of 'gripperOffsets'."
                        )

                assert isinstance(
                    ancestor,
                    (
                        DeckSlotLocation,
                        ModuleLocation,
                        OnLabwareLocation,
                        AddressableAreaLocation,
                    ),
                ), "No gripper offsets for off-deck labware"
                return (
                    direct_parent_offset.dropOffset
                    + self._nominal_gripper_offsets_for_location(
                        location=ancestor
                    ).dropOffset
                    + extra_offset
                )

    # todo(mm, 2024-11-05): This may be incorrect because it does not take the following
    # offsets into account, which *are* taken into account for the actual gripper movement:
    #
    # * The pickup offset in the definition of the parent of the gripped labware.
    # * The "additional offset" or "user offset", e.g. the `pickUpOffset` and `dropOffset`
    #   params in the `moveLabware` command.
    #
    # And this *does* take these extra offsets into account:
    #
    # * The labware's Labware Position Check offset
    #
    # For robustness, we should combine this with `get_gripper_labware_movement_waypoints()`.
    #
    # We should also be more explicit about which offsets act to move the gripper paddles
    # relative to the gripped labware, and which offsets act to change how the gripped
    # labware sits atop its parent. Those have different effects on how far the gripped
    # labware juts beyond the paddles while it's in transit.
    def check_gripper_labware_tip_collision(
        self,
        gripper_homed_position_z: float,
        labware_id: str,
        current_location: OnDeckLabwareLocation,
    ) -> None:
        """Check for potential collision of tips against labware to be lifted."""
        labware_definition = self._labware.get_definition(labware_id)
        pipettes = self._pipettes.get_all()
        for pipette in pipettes:
            # TODO(cb, 2024-01-22): Remove the 1 and 8 channel special case once we are doing X axis validation
            if self._pipettes.get_channels(pipette.id) in [1, 8]:
                return

            tip = self._pipettes.get_attached_tip(pipette.id)
            if tip:
                # NOTE: This call to get_labware_highest_z() uses the labware's LPC offset,
                # which is an inconsistency between this and the actual gripper movement.
                # See the todo comment above this function.
                labware_top_z_when_gripped = gripper_homed_position_z + (
                    self.get_labware_highest_z(labware_id=labware_id)
                    - self.get_labware_grip_point(
                        labware_definition=labware_definition, location=current_location
                    ).z
                )
                # TODO(cb, 2024-01-18): Utilizing the nozzle map and labware X coordinates verify if collisions will occur on the X axis (analysis will use hard coded data to measure from the gripper critical point to the pipette mount)
                if (
                    _PIPETTE_HOMED_POSITION_Z - tip.length
                ) < labware_top_z_when_gripped:
                    raise LabwareMovementNotAllowedError(
                        f"Cannot move labware '{labware_definition.parameters.loadName}' when {int(tip.volume)} µL tips are attached."
                    )
        return

    def _nominal_gripper_offsets_for_location(
        self, location: OnDeckLabwareLocation
    ) -> LabwareMovementOffsetData:
        """Provide the default gripper offset data for the given location type."""
        if isinstance(location, (DeckSlotLocation, AddressableAreaLocation)):
            # TODO we might need a separate type of gripper offset for addressable areas but that also might just
            #   be covered by the drop labware offset/location
            offsets = self._labware.get_deck_default_gripper_offsets()
        elif isinstance(location, ModuleLocation):
            offsets = self._modules.get_default_gripper_offsets(location.moduleId)
        else:
            # Labware is on a labware/adapter
            offsets = self._labware_gripper_offsets(location.labwareId)
        return offsets or LabwareMovementOffsetData(
            pickUpOffset=LabwareOffsetVector(x=0, y=0, z=0),
            dropOffset=LabwareOffsetVector(x=0, y=0, z=0),
        )

    def _labware_gripper_offsets(
        self, labware_id: str
    ) -> Optional[LabwareMovementOffsetData]:
        """Provide the most appropriate gripper offset data for the specified labware.

        We check the types of gripper offsets available for the labware ("default" or slot-based)
        and return the most appropriate one for the overall location of the labware.
        Currently, only module adapters (specifically, the H/S universal flat adapter)
        have non-default offsets that are specific to location of the module on deck,
        so, this code only checks for the presence of those known offsets.
        """
        parent_location = self._labware.get_parent_location(labware_id)
        assert isinstance(
            parent_location,
            (
                DeckSlotLocation,
                ModuleLocation,
                AddressableAreaLocation,
                OnLabwareLocation,
            ),
        ), "No gripper offsets for off-deck labware"

        if isinstance(parent_location, DeckSlotLocation):
            slot_name = parent_location.slotName
        elif isinstance(parent_location, AddressableAreaLocation):
            slot_name = self._addressable_areas.get_addressable_area_base_slot(
                parent_location.addressableAreaName
            )
        else:
            module_loc = self._modules.get_location(parent_location.moduleId)
            slot_name = module_loc.slotName

        slot_based_offset = self._labware.get_child_gripper_offsets(
            labware_id=labware_id, slot_name=slot_name.to_ot3_equivalent()
        )

        return slot_based_offset or self._labware.get_child_gripper_offsets(
            labware_id=labware_id, slot_name=None
        )

    def get_location_sequence(self, labware_id: str) -> LabwareLocationSequence:
        """Provide the LocationSequence specifying the current position of the labware.

        Elements in this sequence contain instance IDs of things. The chain is valid only until the
        labware is moved.
        """
        return self.get_predicted_location_sequence(
            self._labware.get_location(labware_id)
        )

    def get_predicted_location_sequence(
        self,
        labware_location: LabwareLocation,
        labware_pending_load: dict[str, LoadedLabware] | None = None,
    ) -> LabwareLocationSequence:
        """Get the location sequence for this location. Useful for a labware that hasn't been loaded."""
        return self._recurse_labware_location(
            labware_location, [], labware_pending_load or {}
        )

    def _cutout_fixture_location_sequence_from_addressable_area(
        self, addressable_area_name: str
    ) -> OnCutoutFixtureLocationSequenceComponent:
        (
            cutout_id,
            potential_fixtures,
        ) = self._addressable_areas.get_current_potential_cutout_fixtures_for_addressable_area(
            addressable_area_name
        )
        return OnCutoutFixtureLocationSequenceComponent(
            possibleCutoutFixtureIds=sorted(
                [fixture.cutout_fixture_id for fixture in potential_fixtures]
            ),
            cutoutId=cutout_id,
        )

    def _recurse_labware_location_from_aa_component(
        self,
        labware_location: AddressableAreaLocation,
        building: LabwareLocationSequence,
    ) -> LabwareLocationSequence:
        cutout_location = self._cutout_fixture_location_sequence_from_addressable_area(
            labware_location.addressableAreaName
        )
        # If the labware is loaded on an AA that is a module, we want to respect the convention
        # of giving it an OnModuleLocation.
        possible_module = self._modules.get_by_addressable_area(
            labware_location.addressableAreaName
        )
        if possible_module is not None:
            return building + [
                OnAddressableAreaLocationSequenceComponent(
                    addressableAreaName=labware_location.addressableAreaName
                ),
                OnModuleLocationSequenceComponent(moduleId=possible_module.id),
                cutout_location,
            ]
        else:
            return building + [
                OnAddressableAreaLocationSequenceComponent(
                    addressableAreaName=labware_location.addressableAreaName,
                ),
                cutout_location,
            ]

    def _recurse_labware_location_from_module_component(
        self, labware_location: ModuleLocation, building: LabwareLocationSequence
    ) -> LabwareLocationSequence:
        module_id = labware_location.moduleId
        module_aa = self._modules.get_provided_addressable_area(module_id)
        base_location: (
            OnCutoutFixtureLocationSequenceComponent
            | NotOnDeckLocationSequenceComponent
        ) = self._cutout_fixture_location_sequence_from_addressable_area(module_aa)

        if self._modules.get_deck_supports_module_fixtures():
            # On a deck with modules as cutout fixtures, we want, in order,
            # - the addressable area of the module
            # - the module with its module id, which is what clients want
            # - the cutout
            loc = self._modules.get_location(module_id)
            model = self._modules.get_connected_model(module_id)
            module_aa = self._modules.ensure_and_convert_module_fixture_location(
                loc.slotName, model
            )
            return building + [
                OnAddressableAreaLocationSequenceComponent(
                    addressableAreaName=module_aa
                ),
                OnModuleLocationSequenceComponent(moduleId=module_id),
                base_location,
            ]
        else:
            # If the module isn't a cutout fixture, then we want
            # - the module
            # - the addressable area the module is loaded on
            # - the cutout
            location = self._modules.get_location(module_id)
            return building + [
                OnModuleLocationSequenceComponent(moduleId=module_id),
                OnAddressableAreaLocationSequenceComponent(
                    addressableAreaName=location.slotName.value
                ),
                base_location,
            ]

    def _recurse_labware_location_from_stacker_hopper(
        self,
        labware_location: InStackerHopperLocation,
        building: LabwareLocationSequence,
    ) -> LabwareLocationSequence:
        loc = self._modules.get_location(labware_location.moduleId)
        model = self._modules.get_connected_model(labware_location.moduleId)
        module_aa = self._modules.ensure_and_convert_module_fixture_location(
            loc.slotName, model
        )
        cutout_base = self._cutout_fixture_location_sequence_from_addressable_area(
            module_aa
        )
        return building + [labware_location, cutout_base]

    def _recurse_labware_location(
        self,
        labware_location: LabwareLocation,
        building: LabwareLocationSequence,
        labware_pending_load: dict[str, LoadedLabware],
    ) -> LabwareLocationSequence:
        if isinstance(labware_location, AddressableAreaLocation):
            return self._recurse_labware_location_from_aa_component(
                labware_location, building
            )
        elif labware_location_is_off_deck(
            labware_location
        ) or labware_location_is_system(labware_location):
            return building + [
                NotOnDeckLocationSequenceComponent(logicalLocationName=labware_location)
            ]

        elif isinstance(labware_location, OnLabwareLocation):
            labware = self._get_or_default_labware(
                labware_location.labwareId, labware_pending_load
            )
            return self._recurse_labware_location(
                labware.location,
                building
                + [
                    OnLabwareLocationSequenceComponent(
                        labwareId=labware_location.labwareId, lidId=labware.lid_id
                    )
                ],
                labware_pending_load,
            )
        elif isinstance(labware_location, ModuleLocation):
            return self._recurse_labware_location_from_module_component(
                labware_location, building
            )
        elif isinstance(labware_location, DeckSlotLocation):
            return building + [
                OnAddressableAreaLocationSequenceComponent(
                    addressableAreaName=labware_location.slotName.value,
                ),
                self._cutout_fixture_location_sequence_from_addressable_area(
                    labware_location.slotName.value
                ),
            ]
        elif isinstance(labware_location, InStackerHopperLocation):
            return self._recurse_labware_location_from_stacker_hopper(
                labware_location, building
            )
        else:
            _LOG.warn(f"Unhandled labware location kind: {labware_location}")
            return building

    def get_offset_location(
        self, labware_id: str
    ) -> Optional[LabwareOffsetLocationSequence]:
        """Provide the LegacyLabwareOffsetLocation specifying the current position of the labware.

        If the labware is in a location that cannot be specified by a LabwareOffsetLocationSequence
        (for instance, OFF_DECK) then return None.
        """
        parent_location = self._labware.get_location(labware_id)
        return self.get_projected_offset_location(parent_location)

    def get_projected_offset_location(
        self,
        labware_location: LabwareLocation,
        labware_pending_load: dict[str, LoadedLabware] | None = None,
    ) -> Optional[LabwareOffsetLocationSequence]:
        """Get the offset location that a labware loaded into this location would match."""
        return self._recurse_labware_offset_location(
            labware_location, [], labware_pending_load or {}
        )

    def _recurse_labware_offset_location(
        self,
        labware_location: LabwareLocation,
        building: LabwareOffsetLocationSequence,
        labware_pending_load: dict[str, LoadedLabware],
    ) -> LabwareOffsetLocationSequence | None:
        if isinstance(labware_location, DeckSlotLocation):
            return building + [
                OnAddressableAreaOffsetLocationSequenceComponent(
                    addressableAreaName=labware_location.slotName.value
                )
            ]

        elif isinstance(labware_location, ModuleLocation):
            module_id = labware_location.moduleId
            # Allow ModuleNotLoadedError to propagate.
            # Note also that we match based on the module's requested model, not its
            # actual model, to implement robot-server's documented HTTP API semantics.
            module_model = self._modules.get_requested_model(module_id=module_id)

            # If `module_model is None`, it probably means that this module was added by
            # `ProtocolEngine.use_attached_modules()`, instead of an explicit
            # `loadModule` command.
            #
            # This assert should never raise in practice because:
            #   1. `ProtocolEngine.use_attached_modules()` is only used by
            #      robot-server's "stateless command" endpoints, under `/commands`.
            #   2. Those endpoints don't support loading labware, so this code will
            #      never run.
            #
            # Nevertheless, if it does happen somehow, we do NOT want to pass the
            # `None` value along to `LabwareView.find_applicable_labware_offset()`.
            # `None` means something different there, which will cause us to return
            # wrong results.
            assert module_model is not None, (
                "Can't find offsets for labware"
                " that are loaded on modules"
                " that were loaded with ProtocolEngine.use_attached_modules()."
            )

            module_location = self._modules.get_location(module_id=module_id)
            if self._modules.get_deck_supports_module_fixtures():
                module_aa = self._modules.ensure_and_convert_module_fixture_location(
                    module_location.slotName, module_model
                )
            else:
                module_aa = module_location.slotName.value
            return building + [
                OnModuleOffsetLocationSequenceComponent(moduleModel=module_model),
                OnAddressableAreaOffsetLocationSequenceComponent(
                    addressableAreaName=module_aa
                ),
            ]

        elif isinstance(labware_location, OnLabwareLocation):
            parent_labware_id = labware_location.labwareId
            parent_labware = self._get_or_default_labware(
                parent_labware_id, labware_pending_load
            )
            parent_labware_uri = LabwareUri(parent_labware.definitionUri)
            base_location = parent_labware.location
            return self._recurse_labware_offset_location(
                base_location,
                building
                + [
                    OnLabwareOffsetLocationSequenceComponent(
                        labwareUri=parent_labware_uri
                    )
                ],
                labware_pending_load,
            )

        else:  # Off deck
            return None

    def get_liquid_handling_z_change(
        self,
        labware_id: str,
        well_name: str,
        operation_volume: float,
    ) -> float:
        """Get the change in height from a liquid handling operation."""
        initial_handling_height = self.get_meniscus_height(
            labware_id=labware_id, well_name=well_name
        )
        final_height = self.get_well_height_after_liquid_handling(
            labware_id=labware_id,
            well_name=well_name,
            initial_height=initial_handling_height,
            volume=operation_volume,
        )
        # this function is only called by
        # HardwarePipetteHandler::aspirate/dispense while_tracking, and shouldn't
        # be reached in the case of a simulated liquid_probe
        assert not isinstance(initial_handling_height, SimulatedProbeResult)
        assert not isinstance(final_height, SimulatedProbeResult)
        return final_height - initial_handling_height

    def get_well_offset_adjustment(
        self,
        labware_id: str,
        well_name: str,
        well_location: WellLocations,
        well_depth: float,
        operation_volume: Optional[float] = None,
    ) -> LiquidTrackingType:
        """Return a z-axis distance that accounts for well handling height and operation volume.

        Distance is with reference to the well bottom.
        """
        # TODO(pbm, 10-23-24): refactor to smartly reduce height/volume conversions

        initial_handling_height = self.get_well_handling_height(
            labware_id=labware_id,
            well_name=well_name,
            well_location=well_location,
            well_depth=well_depth,
        )
        # if we're tracking a MENISCUS origin, and targeting either the beginning
        #   position of the liquid or doing dynamic tracking, return the initial height
        if (
            well_location.origin == WellOrigin.MENISCUS
            and not well_location.volumeOffset
        ):
            return initial_handling_height
        if isinstance(well_location, PickUpTipWellLocation):
            volume = 0.0
        elif isinstance(well_location.volumeOffset, float):
            volume = well_location.volumeOffset
        elif well_location.volumeOffset == "operationVolume":
            volume = operation_volume or 0.0

        if volume:
            liquid_height_after = self.get_well_height_after_liquid_handling(
                labware_id=labware_id,
                well_name=well_name,
                initial_height=initial_handling_height,
                volume=volume,
            )
            return liquid_height_after
        else:
            return initial_handling_height

    def get_current_well_volume(
        self,
        labware_id: str,
        well_name: str,
    ) -> LiquidTrackingType:
        """Returns most recently updated volume in specified well."""
        last_updated = self._wells.get_last_liquid_update(labware_id, well_name)
        if last_updated is None:
            raise errors.LiquidHeightUnknownError(
                "Must LiquidProbe or LoadLiquid before specifying WellOrigin.MENISCUS."
            )

        well_liquid = self._wells.get_well_liquid_info(
            labware_id=labware_id, well_name=well_name
        )
        if (
            well_liquid.probed_height is not None
            and well_liquid.probed_height.height is not None
            and well_liquid.probed_height.last_probed == last_updated
        ):
            volume = self.get_well_volume_at_height(
                labware_id=labware_id,
                well_name=well_name,
                height=well_liquid.probed_height.height,
            )
            return volume
        elif (
            well_liquid.loaded_volume is not None
            and well_liquid.loaded_volume.volume is not None
            and well_liquid.loaded_volume.last_loaded == last_updated
        ):
            return well_liquid.loaded_volume.volume
        elif (
            well_liquid.probed_volume is not None
            and well_liquid.probed_volume.volume is not None
            and well_liquid.probed_volume.last_probed == last_updated
        ):
            return well_liquid.probed_volume.volume
        else:
            # This should not happen if there was an update but who knows
            raise errors.LiquidVolumeUnknownError(
                f"Unable to find liquid volume despite an update at {last_updated}."
            )

    def get_meniscus_height(
        self,
        labware_id: str,
        well_name: str,
    ) -> LiquidTrackingType:
        """Returns stored meniscus height in specified well."""
        last_updated = self._wells.get_last_liquid_update(labware_id, well_name)
        if last_updated is None:
            raise errors.LiquidHeightUnknownError(
                "Must LiquidProbe or LoadLiquid before specifying WellOrigin.MENISCUS."
            )

        well_liquid = self._wells.get_well_liquid_info(
            labware_id=labware_id, well_name=well_name
        )
        if (
            well_liquid.probed_height is not None
            and well_liquid.probed_height.height is not None
            and well_liquid.probed_height.last_probed == last_updated
        ):
            return well_liquid.probed_height.height
        elif (
            well_liquid.loaded_volume is not None
            and well_liquid.loaded_volume.volume is not None
            and well_liquid.loaded_volume.last_loaded == last_updated
        ):
            return self.get_well_height_at_volume(
                labware_id=labware_id,
                well_name=well_name,
                volume=well_liquid.loaded_volume.volume,
            )
        elif (
            well_liquid.probed_volume is not None
            and well_liquid.probed_volume.volume is not None
            and well_liquid.probed_volume.last_probed == last_updated
        ):
            return self.get_well_height_at_volume(
                labware_id=labware_id,
                well_name=well_name,
                volume=well_liquid.probed_volume.volume,
            )
        else:
            # This should not happen if there was an update but who knows
            raise errors.LiquidHeightUnknownError(
                f"Unable to find liquid height despite an update at {last_updated}."
            )

    def get_well_handling_height(
        self,
        labware_id: str,
        well_name: str,
        well_location: WellLocations,
        well_depth: float,
    ) -> LiquidTrackingType:
        """Return the handling height for a labware well (with reference to the well bottom)."""
        handling_height: LiquidTrackingType = 0.0
        if well_location.origin == WellOrigin.TOP:
            handling_height = float(well_depth)
        elif well_location.origin == WellOrigin.CENTER:
            handling_height = float(well_depth / 2.0)
        elif well_location.origin == WellOrigin.MENISCUS:
            handling_height = self.get_meniscus_height(
                labware_id=labware_id, well_name=well_name
            )
        return handling_height

    def get_well_height_after_liquid_handling(
        self,
        labware_id: str,
        well_name: str,
        initial_height: LiquidTrackingType,
        volume: float,
    ) -> LiquidTrackingType:
        """Return the height of liquid in a labware well after a given volume has been handled.

        This is given an initial handling height, with reference to the well bottom.
        """
        well_geometry = self._labware.get_well_geometry(
            labware_id=labware_id, well_name=well_name
        )
        initial_volume = find_volume_at_well_height(
            target_height=initial_height, well_geometry=well_geometry
        )
        final_volume = initial_volume + volume
        return find_height_at_well_volume(
            target_volume=final_volume, well_geometry=well_geometry
        )

    def get_well_height_after_liquid_handling_no_error(
        self,
        labware_id: str,
        well_name: str,
        initial_height: LiquidTrackingType,
        volume: float,
    ) -> LiquidTrackingType:
        """Return what the height of liquid in a labware well after liquid handling will be.

        This raises no error if the value returned is an invalid physical location, so it should never be
        used for navigation, only for a pre-emptive estimate.
        """
        well_geometry = self._labware.get_well_geometry(
            labware_id=labware_id, well_name=well_name
        )
        initial_volume = find_volume_at_well_height(
            target_height=initial_height, well_geometry=well_geometry
        )
        final_volume = initial_volume + volume
        well_volume = find_height_at_well_volume(
            target_volume=final_volume,
            well_geometry=well_geometry,
            raise_error_if_result_invalid=False,
        )
        return well_volume

    def get_well_height_at_volume(
        self, labware_id: str, well_name: str, volume: LiquidTrackingType
    ) -> LiquidTrackingType:
        """Convert well volume to height."""
        well_geometry = self._labware.get_well_geometry(labware_id, well_name)
        return find_height_at_well_volume(
            target_volume=volume, well_geometry=well_geometry
        )

    def get_well_volume_at_height(
        self,
        labware_id: str,
        well_name: str,
        height: LiquidTrackingType,
    ) -> LiquidTrackingType:
        """Convert well height to volume."""
        well_geometry = self._labware.get_well_geometry(labware_id, well_name)
        return find_volume_at_well_height(
            target_height=height, well_geometry=well_geometry
        )

    def validate_dispense_volume_into_well(
        self,
        labware_id: str,
        well_name: str,
        well_location: WellLocations,
        volume: float,
    ) -> None:
        """Raise InvalidDispenseVolumeError if planned dispense volume will overflow well."""
        well_def = self._labware.get_well_definition(labware_id, well_name)
        well_volumetric_capacity = float(well_def.totalLiquidVolume)
        if well_location.origin == WellOrigin.MENISCUS:
            # TODO(pbm, 10-23-24): refactor to smartly reduce height/volume conversions
            well_geometry = self._labware.get_well_geometry(labware_id, well_name)
            meniscus_height = self.get_meniscus_height(
                labware_id=labware_id, well_name=well_name
            )
            meniscus_volume = find_volume_at_well_height(
                target_height=meniscus_height, well_geometry=well_geometry
            )
            # if meniscus volume is a simulated value, comparisons aren't meaningful
            if isinstance(meniscus_volume, SimulatedProbeResult):
                return
            remaining_volume = well_volumetric_capacity - meniscus_volume
            if volume > remaining_volume:
                raise errors.InvalidDispenseVolumeError(
                    f"Attempting to dispense {volume}µL of liquid into a well that can currently only hold {remaining_volume}µL (well {well_name} in labware_id: {labware_id})"
                )
        else:
            # TODO(pbm, 10-08-24): factor in well (LabwareStore) state volume
            if volume > well_volumetric_capacity:
                raise errors.InvalidDispenseVolumeError(
                    f"Attempting to dispense {volume}µL of liquid into a well that can only hold {well_volumetric_capacity}µL (well {well_name} in labware_id: {labware_id})"
                )

    def get_wells_covered_by_pipette_with_active_well(
        self, labware_id: str, target_well_name: str, pipette_id: str
    ) -> list[str]:
        """Get a flat list of wells that are covered by a pipette when moved to a specified well.

        When you move a pipette in a multichannel configuration  to a specific well - the target well -
        the pipette will operate on other wells as well.

        For instance, a pipette with a COLUMN configuration with well A1 of an SBS standard labware target
        will also "cover", under this definition, wells B1-H1. That same pipette, when C5 is the target well, will "cover"
        wells C5-H5.

        This math only works, and may only be applied, if one of the following is true:
        - The pipette is in a SINGLE configuration
        - The pipette is in a non-SINGLE configuration, and the labware is an SBS-format 96 or 384 well plate (and is so
          marked in its definition's parameters.format key, as 96Standard or 384Standard)

        If all of the following do not apply, regardless of the nozzle configuration of the pipette this function will
        return only the labware covered by the primary well.
        """
        pipette_nozzle_map = self._pipettes.get_nozzle_configuration(pipette_id)
        labware_columns = [
            column for column in self._labware.get_definition(labware_id).ordering
        ]
        try:
            return list(
                wells_covered_by_pipette_configuration(
                    pipette_nozzle_map, target_well_name, labware_columns
                )
            )
        except InvalidStoredData:
            return [target_well_name]

    def get_nozzles_per_well(
        self, labware_id: str, target_well_name: str, pipette_id: str
    ) -> int:
        """Get the number of nozzles that will interact with each well."""
        return nozzles_per_well(
            self._pipettes.get_nozzle_configuration(pipette_id),
            target_well_name,
            self._labware.get_definition(labware_id).ordering,
        )

    def get_height_of_labware_stack(
        self, definitions: list[LabwareDefinition]
    ) -> float:
        """Get the overall height of a stack of labware listed by definition in top-first order."""
        if len(definitions) == 0:
            return 0
        if len(definitions) == 1:
            return definitions[0].dimensions.zDimension
        total_height = 0.0
        upper_def: LabwareDefinition = definitions[0]
        for lower_def in definitions[1:]:
            overlap = self._labware.get_labware_overlap_offsets(
                upper_def, lower_def.parameters.loadName
            ).z
            total_height += upper_def.dimensions.zDimension - overlap
            upper_def = lower_def
        return total_height + upper_def.dimensions.zDimension

    def get_height_of_stacker_labware_pool(self, module_id: str) -> float:
        """Get the overall height of a stack of labware in a Stacker module."""
        stacker = self._modules.get_flex_stacker_substate(module_id)
        pool_list = stacker.get_pool_definition_ordered_list()
        if not pool_list:
            return 0.0
        return self.get_height_of_labware_stack(pool_list)

    def _get_or_default_labware(
        self, labware_id: str, pending_labware: dict[str, LoadedLabware]
    ) -> LoadedLabware:
        try:
            return self._labware.get(labware_id)
        except LabwareNotLoadedError as lnle:
            try:
                return pending_labware[labware_id]
            except KeyError as ke:
                raise lnle from ke



"""Command models for opening a gripper jaw."""
from __future__ import annotations
from typing import Literal, Type, Optional
from opentrons.hardware_control import HardwareControlAPI
from opentrons.protocol_engine.resources import ensure_ot3_hardware

from pydantic import BaseModel

from ..command import (
    AbstractCommandImpl,
    BaseCommand,
    BaseCommandCreate,
    SuccessData,
)
from opentrons.protocol_engine.errors.error_occurrence import ErrorOccurrence


openGripperJawCommandType = Literal["robot/openGripperJaw"]


class openGripperJawParams(BaseModel):
    """Payload required to release a gripper."""

    pass


class openGripperJawResult(BaseModel):
    """Result data from the execution of a openGripperJaw command."""

    pass


class openGripperJawImplementation(
    AbstractCommandImpl[openGripperJawParams, SuccessData[openGripperJawResult]]
):
    """openGripperJaw command implementation."""

    def __init__(
        self,
        hardware_api: HardwareControlAPI,
        **kwargs: object,
    ) -> None:
        self._hardware_api = hardware_api

    async def execute(
        self, params: openGripperJawParams
    ) -> SuccessData[openGripperJawResult]:
        """Release the gripper."""
        ot3_hardware_api = ensure_ot3_hardware(self._hardware_api)

        await ot3_hardware_api.home_gripper_jaw()
        return SuccessData(
            public=openGripperJawResult(),
        )


class openGripperJaw(
    BaseCommand[openGripperJawParams, openGripperJawResult, ErrorOccurrence]
):
    """openGripperJaw command model."""

    commandType: openGripperJawCommandType = "robot/openGripperJaw"
    params: openGripperJawParams
    result: Optional[openGripperJawResult] = None

    _ImplementationCls: Type[
        openGripperJawImplementation
    ] = openGripperJawImplementation


class openGripperJawCreate(BaseCommandCreate[openGripperJawParams]):
    """openGripperJaw command request model."""

    commandType: openGripperJawCommandType = "robot/openGripperJaw"
    params: openGripperJawParams

    _CommandCls: Type[openGripperJaw] = openGripperJaw



"""Implementation, request models, and response models for the load module command."""
from __future__ import annotations
from typing import TYPE_CHECKING, Optional, Type, Any
from typing_extensions import Literal
from pydantic import BaseModel, Field
from pydantic.json_schema import SkipJsonSchema

from opentrons.protocol_engine.state.update_types import StateUpdate

from .command import AbstractCommandImpl, BaseCommand, BaseCommandCreate, SuccessData
from ..errors.error_occurrence import ErrorOccurrence
from ..types import (
    DeckSlotLocation,
    AddressableAreaLocation,
    ModuleType,
    ModuleModel,
    ModuleDefinition,
)
from opentrons.types import DeckSlotName

from opentrons.protocol_engine.resources import deck_configuration_provider


if TYPE_CHECKING:
    from ..state.state import StateView
    from ..execution import EquipmentHandler


LoadModuleCommandType = Literal["loadModule"]


def _remove_default(s: dict[str, Any]) -> None:
    s.pop("default", None)


class LoadModuleParams(BaseModel):
    """Payload required to load a module."""

    model: ModuleModel = Field(
        ...,
        description=(
            "The model name of the module to load."
            "\n\n"
            "Protocol Engine will look for a connected module that either"
            " exactly matches this one, or is compatible."
            "\n\n"
            " For example, if you request a `temperatureModuleV1` here,"
            " Protocol Engine might load a `temperatureModuleV1` or a `temperatureModuleV2`."
            "\n\n"
            " The model that it finds connected will be available through `result.model`."
        ),
    )

    # Note: Our assumption here that a module's position can be boiled down to a
    # single deck slot precludes loading a Thermocycler in its special "shifted slightly
    # to the left" position. This is okay for now because neither the Python Protocol
    # API nor Protocol Designer attempt to support it, either.
    location: DeckSlotLocation = Field(
        ...,
        description=(
            "The location into which this module should be loaded."
            "\n\n"
            "For the Thermocycler Module, which occupies multiple deck slots,"
            " this should be the front-most occupied slot (normally slot 7)."
        ),
    )

    moduleId: str | SkipJsonSchema[None] = Field(
        None,
        description=(
            "An optional ID to assign to this module."
            " If None, an ID will be generated."
        ),
        json_schema_extra=_remove_default,
    )


class LoadModuleResult(BaseModel):
    """The results of loading a module."""

    moduleId: str = Field(
        description="An ID to reference this module in subsequent commands."
    )

    # TODO(mm, 2023-04-13): Remove this field. Jira RSS-221.
    definition: ModuleDefinition = Field(
        json_schema_extra={"deprecated": True},
        description=(
            "The definition of the connected module."
            " This field is an implementation detail. We might change or remove it without warning."
            " Do not access it or rely on it being present."
        ),
    )

    model: ModuleModel = Field(
        ...,
        description=(
            "The hardware model of the connected module."
            " This can be different from the exact model that this command requested."
            " See `params.model`."
            "\n\n"
            "This field is only meaningful in the run's actual execution,"
            " not in the protocol's analysis."
            " In analysis, it will be an arbitrary placeholder."
        ),
    )

    serialNumber: Optional[str] = Field(
        None,
        description="Hardware serial number of the connected module. "
        "Will be `None` if a module is not electrically connected to the robot (like the Magnetic Block).",
    )


class LoadModuleImplementation(
    AbstractCommandImpl[LoadModuleParams, SuccessData[LoadModuleResult]]
):
    """The implementation of the load module command."""

    def __init__(
        self, equipment: EquipmentHandler, state_view: StateView, **kwargs: object
    ) -> None:
        self._equipment = equipment
        self._state_view = state_view

    async def execute(self, params: LoadModuleParams) -> SuccessData[LoadModuleResult]:
        """Check that the requested module is attached and assign its identifier."""
        state_update = StateUpdate()

        module_type = params.model.as_type()
        self._ensure_module_location(params.location.slotName, module_type)

        if self._state_view.modules.get_deck_supports_module_fixtures():
            addressable_area_module_reference = (
                self._state_view.modules.ensure_and_convert_module_fixture_location(
                    deck_slot=params.location.slotName,
                    model=params.model,
                )
            )
        else:
            addressable_area_module_reference = params.location.slotName.id
            state_update.set_addressable_area_used(
                addressable_area_name=addressable_area_module_reference
            )

        self._state_view.addressable_areas.raise_if_area_not_in_deck_configuration(
            addressable_area_module_reference
        )

        self._state_view.geometry.ensure_location_not_occupied(
            params.location, addressable_area_module_reference
        )

        if params.model == ModuleModel.MAGNETIC_BLOCK_V1:
            loaded_module = await self._equipment.load_magnetic_block(
                model=params.model,
                location=AddressableAreaLocation(
                    addressableAreaName=addressable_area_module_reference
                ),
                module_id=params.moduleId,
            )
        else:
            loaded_module = await self._equipment.load_module(
                model=params.model,
                location=AddressableAreaLocation(
                    addressableAreaName=addressable_area_module_reference
                ),
                module_id=params.moduleId,
            )

        return SuccessData(
            public=LoadModuleResult(
                moduleId=loaded_module.module_id,
                serialNumber=loaded_module.serial_number,
                model=loaded_module.definition.model,
                definition=loaded_module.definition,
            ),
            state_update=state_update,
        )

    def _ensure_module_location(
        self, slot: DeckSlotName, module_type: ModuleType
    ) -> None:
        # todo(mm, 2024-12-03): Theoretically, we should be able to deal with
        # addressable areas and deck configurations the same way between OT-2 and Flex.
        # Can this be simplified?
        if self._state_view.config.robot_type == "OT-2 Standard":
            slot_def = self._state_view.addressable_areas.get_slot_definition(slot.id)
            compatible_modules = slot_def["compatibleModuleTypes"]
            if module_type.value not in compatible_modules:
                raise ValueError(
                    f"A {module_type.value} cannot be loaded into slot {slot}"
                )
        else:
            cutout_fixture_id = ModuleType.to_module_fixture_id(module_type)
            module_fixture = deck_configuration_provider.get_cutout_fixture(
                cutout_fixture_id,
                self._state_view.labware.get_deck_definition(),
            )
            cutout_id = (
                self._state_view.addressable_areas.get_cutout_id_by_deck_slot_name(slot)
            )
            if cutout_id not in module_fixture["mayMountTo"]:
                raise ValueError(
                    f"A {module_type.value} cannot be loaded into slot {slot}"
                )


class LoadModule(BaseCommand[LoadModuleParams, LoadModuleResult, ErrorOccurrence]):
    """The model for a load module command."""

    commandType: LoadModuleCommandType = "loadModule"
    params: LoadModuleParams
    result: Optional[LoadModuleResult] = None

    _ImplementationCls: Type[LoadModuleImplementation] = LoadModuleImplementation


class LoadModuleCreate(BaseCommandCreate[LoadModuleParams]):
    """The model for a creation request for a load module command."""

    commandType: LoadModuleCommandType = "loadModule"
    params: LoadModuleParams

    _CommandCls: Type[LoadModule] = LoadModule



"""Helpers for commands that alter the position of labware."""

from pydantic import BaseModel, Field

from ..types import LabwareLocationSequence


class LabwareHandlingResultMixin(BaseModel):
    """A result for commands that create a labware entity."""

    labwareId: str = Field(..., description="The id of the labware.")
    locationSequence: LabwareLocationSequence | None = Field(
        None,
        description="The full location down to the deck on which this labware exists.",
    )


class LabwarePositionResultMixin(LabwareHandlingResultMixin):
    """A result for commands that create an offsetable labware entity."""

    offsetId: str | None = Field(
        None,
        description="An ID referencing the labware offset that will apply to this labware in this location.",
    )



"""Utility functions for transfer_liquid, consolidate_liquid and distribute_liquid"""
from __future__ import annotations

from typing import Literal, Sequence, List, Optional, TYPE_CHECKING
from dataclasses import dataclass

from opentrons.protocol_engine.errors import LiquidHeightUnknownError
from opentrons.protocol_engine.state._well_math import (
    wells_covered_by_pipette_configuration,
)
from opentrons.types import NozzleMapInterface, NozzleConfigurationType

if TYPE_CHECKING:
    from logging import Logger
    from opentrons.types import Location
    from opentrons.protocol_api.core.engine import WellCore
    from opentrons.protocol_api.labware import Well, Labware


@dataclass
class LocationCheckDescriptors:
    location_type: Literal["submerge start", "retract end"]
    pipetting_action: Literal["aspirate", "dispense"]


def raise_if_location_inside_liquid(
    location: Location,
    well_location: Location,
    well_core: WellCore,
    location_check_descriptors: LocationCheckDescriptors,
    logger: Logger,
) -> None:
    """Raise an error if the location in question would be inside the liquid.

    This checker will raise an error if:
    - the location in question is below the target well location during aspirate/dispense or,
    - if we can find the liquid height AND the location in question is below this height.
      If we can't find the liquid height, then we simply log a warning and no error is raised.
    """
    if location.point.z < well_location.point.z:
        raise RuntimeError(
            f"Received {location_check_descriptors.location_type} location of {location}"
            f" and {location_check_descriptors.pipetting_action} location of {well_location}."
            f" {location_check_descriptors.location_type.capitalize()} location z should not be lower"
            f" than the {location_check_descriptors.pipetting_action} location z."
        )
    try:
        liquid_height_from_bottom = well_core.current_liquid_height()
    except LiquidHeightUnknownError:
        liquid_height_from_bottom = None
    if isinstance(liquid_height_from_bottom, (int, float)):
        if liquid_height_from_bottom + well_core.get_bottom(0).z > location.point.z:
            raise RuntimeError(
                f"{location_check_descriptors.location_type.capitalize()} location {location} is"
                f" inside the liquid in well {well_core.get_display_name()} when it should be outside"
                f"(above) the liquid."
            )
    else:
        # We could raise an error here but that would restrict the use of
        # liquid classes-based transfer to only when LPD is enabled or when liquids are
        # loaded in protocols using `load_liquid`. This can be quite restrictive
        # so we will not raise but just log a warning.
        logger.warning(
            f"Could not verify height of liquid in well {well_core.get_display_name()}, either"
            f" because the liquid in this well has not been probed or because"
            f" liquid was not loaded in this well using `load_liquid`."
            f" Proceeding without verifying if {location_check_descriptors.location_type}"
            f" location is outside the liquid."
        )


def group_wells_for_multi_channel_transfer(
    targets: Sequence[Well],
    nozzle_map: NozzleMapInterface,
) -> List[Well]:
    """Takes a list of wells and a nozzle map and returns a list of target wells to address every well given

    This currently only supports 8-tip columns, 12-tip rows and full 96-channel configurations,
    and only is used for 96 and 384 well plates. This assumes the wells are being given in a
    contiguous order (or every other for 384), and will raise if a well is found that does not overlap
    with the first target well given for a sequence, or if not all wells are given for that sequence.
    """
    configuration = nozzle_map.configuration
    active_nozzles = nozzle_map.tip_count

    if (
        (
            (
                configuration == NozzleConfigurationType.COLUMN
                or configuration == NozzleConfigurationType.FULL
            )
            and active_nozzles == 8
        )
        or (configuration == NozzleConfigurationType.ROW and active_nozzles == 12)
        or active_nozzles == 96
    ):
        return _group_wells_for_nozzle_configuration(list(targets), nozzle_map)
    else:
        raise ValueError("Unsupported tip configuration for well grouping")


def _group_wells_for_nozzle_configuration(  # noqa: C901
    targets: List[Well], nozzle_map: NozzleMapInterface
) -> List[Well]:
    """Groups wells together for a column, row, or full 96 configuration and returns a reduced list of target wells."""
    grouped_wells = []
    active_wells_covered: List[str] = []
    active_labware: Optional[Labware] = None
    alternate_384_well_coverage_count = 0
    labware_format: Optional[str] = None

    # We are assuming the wells are ordered A1, B1, C1... A2, B2, C2..., for columns and
    # A1, A2, A3... B1, B2, B3 for rows. So if the active nozzle is on H row/12 column,
    # reverse the list so the correct primary nozzle is chosen
    reverse_lookup = (
        nozzle_map.starting_nozzle == "H12"
        or (
            nozzle_map.configuration == NozzleConfigurationType.COLUMN
            and nozzle_map.starting_nozzle == "H1"
        )
        or (
            nozzle_map.configuration == NozzleConfigurationType.ROW
            and nozzle_map.starting_nozzle == "A12"
        )
    )
    if reverse_lookup:
        targets.reverse()

    for well in targets:
        # If we have wells that are covered by the pipette's nozzles while primary nozzle is over
        # a target well that aren't accounted for, check if the current well is in that list
        if active_wells_covered:
            if well.parent != active_labware:
                raise ValueError(
                    "Could not resolve wells provided to pipette's nozzle configuration. "
                    "Please ensure wells are ordered to match pipette's nozzle layout."
                )

            if well.well_name in active_wells_covered:
                active_wells_covered.remove(well.well_name)
            # If it's a 384 well plate, contiguous wells are not covered by the pipette targeting the
            # initial target well. To support these kinds of transfers given a list of contiguous wells,
            # allow another target well (or up to 4 total for a full 96-tip config) and add those wells
            # to the list of covered wells
            elif labware_format == "384Standard" and (
                alternate_384_well_coverage_count == 0
                or (
                    nozzle_map.tip_count == 96 and alternate_384_well_coverage_count < 3
                )
            ):
                active_wells_covered.extend(
                    list(
                        wells_covered_by_pipette_configuration(
                            nozzle_map,  # type: ignore[arg-type]
                            well.well_name,
                            labware_wells_by_column=[
                                [labware_well.well_name for labware_well in column]
                                for column in well.parent.columns()
                            ],
                        )
                    )
                )
                active_wells_covered.remove(well.well_name)
                grouped_wells.append(well)
                alternate_384_well_coverage_count += 1
            else:
                raise ValueError(
                    "Could not resolve wells provided to pipette's nozzle configuration. "
                    "Please ensure wells are ordered to match pipette's nozzle layout."
                )
        # If we have no active wells covered to account for, add a new target well and list of covered wells to check
        else:
            # If the labware is not a 96 or 384 well plate, add this well to the final result and move on to the next
            labware_format = well.parent.parameters["format"]
            if labware_format != "96Standard" and labware_format != "384Standard":
                grouped_wells.append(well)
                continue

            active_wells_covered = list(
                wells_covered_by_pipette_configuration(
                    nozzle_map,  # type: ignore[arg-type]
                    well.well_name,
                    labware_wells_by_column=[
                        [labware_well.well_name for labware_well in column]
                        for column in well.parent.columns()
                    ],
                )
            )
            active_wells_covered.remove(well.well_name)
            grouped_wells.append(well)
            active_labware = well.parent
            alternate_384_well_coverage_count = 0

    if active_wells_covered:
        raise ValueError(
            "Could not target all wells provided without aspirating or dispensing from other wells. "
            f"Other wells that would be targeted: {active_wells_covered}"
        )

    # If we reversed the lookup of wells, reverse the grouped wells we will return
    if reverse_lookup:
        grouped_wells.reverse()

    return grouped_wells
